import asttrs
import attr
import pytest

from airfly.utils import qualname


class TestLiteral:
    @classmethod
    def setup_class(cls):
        from airfly.model.v1 import Literal

        cls.Literal = Literal

    @classmethod
    def teardown_class(cls):
        pass

    def setup_method(self, method):
        pass

    def teardown_method(self, method):
        pass

    @pytest.mark.parametrize(
        "expected", ["MyClass(*args, **kwargs)", "my_wrapper(some_function)"]
    )
    def test_repr(self, expected):

        assert repr(self.Literal(expected)) == expected

    def test__to_ast(self):

        assert self.Literal("any expression")._to_ast() == asttrs.Name(
            id="any expression", ctx=asttrs.Load()
        )

    def test__dep_ast(self):

        assert (
            #
            self.Literal("any expression", deps=self.Literal)._dep_ast()[0]
            == asttrs.ImportFrom(
                module=self.Literal.__module__, names=[asttrs.alias(name="Literal")]
            )
        )


class TestParam:
    @classmethod
    def setup_class(cls):
        from airfly.model.v1 import Param

        cls.Param = Param

    @classmethod
    def teardown_class(cls):
        pass

    def setup_method(self, method):
        pass

    def teardown_method(self, method):
        pass

    @pytest.mark.parametrize(
        "obj,expected",
        [
            (123, asttrs.Constant(value=123)),
            ("abc", asttrs.Constant(value="abc")),
            (
                [123],
                asttrs.List(
                    elts=[asttrs.Constant(value=123)],
                    ctx=asttrs.Load(),
                ),
            ),
            (
                dict(abc=123),
                asttrs.Dict(
                    keys=[asttrs.Constant(value="abc")],
                    values=[asttrs.Constant(value=123)],
                ),
            ),
        ],
    )
    def test__target_ast(self, obj, expected):

        self.Param(obj)._target_ast() == expected

    def test__target_ast_other_cases(self):
        from airfly.model import Literal

        assert self.Param(self.Param)._target_ast() == asttrs.Name(
            id="Param", ctx=asttrs.Load()
        )

        assert self.Param([self.Param])._target_ast() == asttrs.List(
            elts=[asttrs.Name(id="Param", ctx=asttrs.Load())], ctx=asttrs.Load()
        )

        assert self.Param(self.setup_class)._target_ast() == asttrs.Name(
            id="TestParam.setup_class", ctx=asttrs.Load()
        )

        assert self.Param([self.setup_class])._target_ast() == asttrs.List(
            elts=[asttrs.Name(id="TestParam.setup_class", ctx=asttrs.Load())],
            ctx=asttrs.Load(),
        )

        # Literal cases
        assert self.Param(Literal("any expression"))._target_ast() == asttrs.Name(
            id="any expression", ctx=asttrs.Load()
        )

        with pytest.raises(TypeError):
            self.Param(lambda: True)._target_ast()

        with pytest.raises(TypeError):
            self.Param(object())._target_ast()

    @pytest.mark.parametrize(
        "obj", [123, "abc", [123], {"abc"}, dict(abc=123), object()]
    )
    def test__dep_ast(self, obj):

        assert self.Param(obj)._dep_ast() == []

    def test__dep_ast_other_cases(self):
        from airfly.model import Literal

        assert (
            self.Param(Literal("Literal('any expression')", deps=Literal))._dep_ast()
            == self.Param(Literal)._dep_ast()
        )

        assert (
            #
            self.Param(
                Literal(
                    "Literal('any expression')", deps=Literal("any import expression")
                )
            )._dep_ast()
            == asttrs.Name(id="any import expression", ctx=asttrs.Load())
        )

    def test__dep_ast_other_cases(self):

        assert self.Param(self.Param)._dep_ast()[0] == asttrs.ImportFrom(
            module=self.Param.__module__, names=[asttrs.alias(name=self.Param.__name__)]
        )

        assert self.Param(self.Param, alias="Param_1")._dep_ast()[
            0
        ] == asttrs.ImportFrom(
            module=self.Param.__module__,
            names=[asttrs.alias(name=self.Param.__name__, asname="Param_1")],
        )


class TestParamContext:
    @classmethod
    def setup_class(cls):
        from airfly.model.v1 import ParamContext

        cls.ParamContext = ParamContext

    @classmethod
    def teardown_class(cls):
        pass

    def setup_method(self, method):
        pass

    def teardown_method(self, method):
        pass

    def test_get(self):

        assert self.ParamContext.get(self.ParamContext).target == self.ParamContext

    def test__get(self):

        ctx = self.ParamContext()
        par1 = ctx._get(self.ParamContext)

        key1 = qualname(self.ParamContext)
        assert key1 in ctx.params
        assert ctx.params[key1] == par1

        _ = ctx._get([self.ParamContext])
        assert len(ctx.params) == 1

        _ = ctx._get(dict(param_ctx=self.ParamContext))
        assert len(ctx.params) == 1

    def test__find_alias_without_conflict(self):

        conflicts = {"key": 0}
        ctx = self.ParamContext()
        ctx.conflicts = conflicts

        assert ctx._find_alias_without_conflict("key") == "key_1"
        assert conflicts["key"] == 1

        conflicts["key_2"] = 0
        assert ctx._find_alias_without_conflict("key") == "key_2_1"
        assert conflicts["key_2"] == 1


class TestTaskAttribute:
    @classmethod
    def setup_class(cls):
        from airfly.model.v1 import TaskAttribute

        assert TaskAttribute

    @classmethod
    def teardown_class(cls):
        pass

    def setup_method(self, method):
        pass

    def teardown_method(self, method):
        pass


class TestTask:
    @classmethod
    def setup_class(cls):
        from airfly.model.v1 import Task

        cls.Task = Task

    @classmethod
    def teardown_class(cls):
        pass

    def setup_method(self, method):
        pass

    def teardown_method(self, method):
        pass

    def test__get_taskid(self):

        assert self.Task._get_taskid(self.Task) == qualname(self.Task)

        class MyTask:
            @classmethod
            def _get_taskid(cls):
                return "my_task"

        assert self.Task._get_taskid(MyTask) == "my_task"

    def test__get_attributes(self):
        from airfly.model.v1 import TaskAttribute
        from airfly.utils import immutable

        # Base case
        class Task1: ...

        class Task2: ...

        class Task3: ...

        task_attr = immutable(TaskAttribute)(
            op_class="MyOperator",
            op_params=dict(abc=123),
            upstream=[Task1, Task2],
            downstream=Task3,
        )

        class Task4:
            op_class = task_attr.op_class

            @property
            def op_params(self):
                return task_attr.op_params

            @property
            def upstream(self):
                return task_attr.upstream

            downstream = task_attr.downstream

        cache = self.Task._get_attributes(Task4)

        assert self.Task._get_attributes(Task4) is cache

        for a in attr.fields(task_attr.__class__):
            assert getattr(cache, a.name) == getattr(task_attr, a.name)

        # Test attribute inherience

        parent_module = "parent_module"
        parent_params = None
        parent_upstream = Task1
        parent_downstream = None
        child_module = "child_module"
        child_params = dict(child=2)
        child_upstream = Task2
        child_downstream = Task3

        class Task5:
            op_class = "ParamOperator"

            # defined class-var
            op_module = parent_module

            # None class-var
            op_params = parent_params

            # defined property
            @property
            def upstream(self):
                return parent_upstream

            # None property
            @property
            def downstream(self):
                return parent_downstream

        class Task6(Task5):
            """Implement nothing
            Expect inherient attributes from parent for all cases.
            """

            ...

        attr6 = self.Task._get_attributes(Task6)
        assert attr6.op_module == parent_module
        assert attr6.op_params is parent_params is None
        assert attr6.upstream == parent_upstream
        assert attr6.downstream is parent_downstream is None

        class Task7(Task5):
            """Define(override) all attributes with class-var"""

            op_module = child_module
            op_params = child_params
            upstream = child_upstream
            downstream = child_downstream

        attr7 = self.Task._get_attributes(Task7)
        assert attr7.op_module == child_module
        assert attr7.op_params == child_params
        assert attr7.upstream == child_upstream
        assert attr7.downstream == child_downstream

        class Task8(Task5):
            """Reset defined attributs to None with class-var"""

            op_module = None
            upstream = None

        attr8 = self.Task._get_attributes(Task8)
        assert attr8.op_module is None
        assert attr8.upstream is None

        class Task9(Task5):
            """Define(override) all attributes with property"""

            @property
            def op_module(self):
                return child_module

            @property
            def op_params(self):
                return child_params

            @property
            def upstream(self):
                return child_upstream

            @property
            def downstream(self):
                return child_downstream

        attr9 = self.Task._get_attributes(Task9)
        assert attr9.op_module == child_module
        assert attr9.op_params == child_params
        assert attr9.upstream == child_upstream
        assert attr9.downstream == child_downstream

        class Task10(Task5):
            """Reset defined attributs to None with property"""

            @property
            def op_module(self):
                return None

            @property
            def upstream(self):
                return None

        attr10 = self.Task._get_attributes(Task10)

        assert attr10.op_module is None
        assert attr10.upstream is None

    def test__to_varname(self):
        assert self.Task._to_varname(self.Task) == qualname(self.Task).replace(".", "_")

        class MyTask:
            @classmethod
            def _to_varname(cls):
                return "my_var"

        assert self.Task._to_varname(MyTask) == "my_var"

    def test__collect_dep_ast(self):

        class Task1:
            op_class = "BashOperator"

        class Task2:
            op_class = "BashOperator"
            op_params = dict(my_dep=self.Task)

        assert self.Task._collect_dep_ast(Task1)[0] == asttrs.ImportFrom(
            module="airflow.operators.bash",
            names=[asttrs.alias(name="BashOperator")],
        )

        assert asttrs.ImportFrom(
            module="airfly.model.v1", names=[asttrs.alias(name="Task")]
        ) in self.Task._collect_dep_ast(Task2)

    def test__to_ast(self):

        class Task:
            op_class = "BashOperator"

        assert isinstance(self.Task._to_ast(Task), asttrs.Assign)

    def test__resolve_operator(self):
        from airfly._vendor.airflow.operators.bash import BashOperator
        from airfly._vendor.airflow.providers.google.cloud.operators.dataproc import (
            DataprocCreateClusterOperator,
        )

        class Task1:
            op_class = "BashOperator"

        class Task2:
            op_class = BashOperator

        class Task3:
            op_class = DataprocCreateClusterOperator

        class Task4:
            op_class = "NonExistentOperator"

        class MyOperator: ...

        class Task5:
            op_class = MyOperator

        class Task6:
            op_class = "DataprocCreateClusterOperator"

        class Task7:
            op_class = "DataprocCreateClusterOperator"
            op_module = "google"

        assert self.Task._resolve_operator(Task1) == BashOperator
        assert self.Task._resolve_operator(Task2) == BashOperator
        assert self.Task._resolve_operator(Task3) == DataprocCreateClusterOperator

        with pytest.raises(ValueError):
            self.Task._resolve_operator(Task4)

        with pytest.raises(TypeError):
            self.Task._resolve_operator(Task5)

        with pytest.raises(ValueError):
            self.Task._resolve_operator(Task6)

        assert self.Task._resolve_operator(Task7) == DataprocCreateClusterOperator

    def test__collect_op_annotations(self):
        pass

    @pytest.mark.parametrize(
        "op_name",
        ["BashOperator", "PythonOperator", "KubernetesPodOperator"],
    )
    def test__is_builtin_op(self, op_name):

        class Task(self.Task):
            op_class = op_name

        op = self.Task._resolve_operator(Task)

        assert self.Task._is_builtin_op(op)

    def test__is_private_op(self, mocker):

        assert not self.Task._is_private_op(mocker.MagicMock())


class TestAirFly:
    @classmethod
    def setup_class(cls):
        from airfly.model.v1 import AirFly

        assert AirFly

    @classmethod
    def teardown_class(cls):
        pass

    def setup_method(self, method):
        pass

    def teardown_method(self, method):
        pass


class TestTaskPair:
    @classmethod
    def setup_class(cls):
        from airfly.model.v1 import TaskPair

        cls.TaskPair = TaskPair

    @classmethod
    def teardown_class(cls):
        pass

    def setup_method(self, method):
        pass

    def teardown_method(self, method):
        pass

    def test__to_ast(self, mocker):

        mock_Task = mocker.patch("airfly.model.v1.Task")

        class Task1: ...

        class Task2: ...

        pair = self.TaskPair(up=Task1, down=Task2)

        _ast = pair._to_ast()

        assert isinstance(_ast, asttrs.Expr)
        assert isinstance(_ast.value, asttrs.BinOp)
        assert isinstance(_ast.value.op, asttrs.RShift)

        assert mock_Task._to_varname.call_count == 2


class TestTaskTree:
    @classmethod
    def setup_class(cls):
        from airfly.model.v1 import TaskTree

        cls.TaskTree = TaskTree

    @classmethod
    def teardown_class(cls):
        pass

    def setup_method(self, method):
        pass

    def teardown_method(self, method):
        pass

    def test__create_dag(self):
        pass

    def test_from_module(self):
        pass

    def test__should_exclude(self):

        class Foo: ...

        assert self.TaskTree._should_exclude(Foo, "Foo")
        assert self.TaskTree._should_exclude(Foo, "_gutt")
        assert not self.TaskTree._should_exclude(Foo, "Bar")

    def test__collect_taskclass(self):
        pass

    def test__collect_taskpairs(self):

        class Task:
            op_class = "BashOperator"

        class Task1(Task): ...

        class Task2(Task):
            upstream = Task1

        class Task3(Task):
            downstream = Task2

        class Task4(Task):
            upstream = [Task1, Task2]
            downstream = Task3

        assert len(list(self.TaskTree._collect_taskpairs({Task4}, Task))) == 3
        assert len(list(self.TaskTree._collect_taskpairs({Task1, Task4}, Task))) == 3
        assert len(list(self.TaskTree._collect_taskpairs({Task3, Task4}, Task))) == 4
        assert (
            len(list(self.TaskTree._collect_taskpairs({Task2, Task3, Task4}, Task)))
            == 5
        )

    def test_to_dag(self):
        pass

    def test__to_ast(self):
        pass

    def test__build_header(self):
        import airfly

        tree = self.TaskTree(taskset=set(), taskpairs=set())
        header = tree._build_header()[0]

        assert isinstance(header, asttrs.Comment)
        assert airfly.__version__ in header.body

    def test__build_imports(self):

        tree = self.TaskTree(taskset={}, taskpairs=set())
        _imports = tree._build_imports()

        assert (
            asttrs.ImportFrom(module="airflow.models", names=[asttrs.alias(name="DAG")])
            in _imports
        )

    def test__build_includes(self, mocker):
        mock_insert_from_pyfile = mocker.patch(
            "airfly.model.v1.TaskTree._insert_from_pyfile"
        )
        tree = self.TaskTree(taskset=set(), taskpairs=set())

        assert tree._build_includes() == []

        includes = ["path1", "path2", "path3"]
        _ = tree._build_includes(includes)

        assert mock_insert_from_pyfile.call_count == len(includes)

    def test__insert_from_pyfile(self):
        pass

    def test__build_dag_context(self, mocker):
        mock_build_dag_body = mocker.patch(
            "airfly.model.v1.TaskTree._build_dag_body", return_value=[asttrs.Pass()]
        )

        tree = self.TaskTree(taskset=set(), taskpairs=set())

        dag_name = "my_dag"
        dag_params = ("a_python_file.py", "a_var_name")
        _ctx_mgr = tree._build_dag_context(dag_name, dag_params)[0]

        assert mock_build_dag_body.call_count == 1
        assert isinstance(_ctx_mgr, asttrs.With)

        _withitem = _ctx_mgr.items[0]
        assert isinstance(_withitem.context_expr, asttrs.Call)

        _call = _withitem.context_expr
        assert isinstance(_call.args[0], asttrs.Constant)
        assert _call.args[0].value == dag_name

        assert isinstance(_call.keywords[0].value, asttrs.Name)
        assert _call.keywords[0].value.id == dag_params[1]

    def test__build_dag_body(self, mocker):
        mock_TaskPair = mocker.patch("airfly.model.v1.TaskPair")
        mock_Task = mocker.patch("airfly.model.v1.Task")

        class Task1: ...

        class Task2: ...

        class Task3: ...

        taskset = {Task1, Task2, Task3}
        taskpairs = {
            mock_TaskPair(up=Task1, down=Task2),
            mock_TaskPair(up=Task1, down=Task3),
            mock_TaskPair(up=Task2, down=Task3),
        }

        empty_tree = self.TaskTree(taskset=set(), taskpairs=set())

        body = empty_tree._build_dag_body()
        assert len(body) == 1 and body[0] == asttrs.Pass()

        tree = self.TaskTree(taskset=taskset, taskpairs=taskpairs)

        _ = tree._build_dag_body()

        assert mock_Task._to_ast.call_count == len(taskset)
        assert mock_Task._collect_dep_ast.call_count == len(taskset)

        for pair in taskpairs:
            assert pair._to_ast.call_count == 1

    def test__build_task_group(self):
        pass


class TestTaskGroup:
    @classmethod
    def setup_class(cls):
        from airfly.model.v1 import TaskGroup

        assert TaskGroup

    @classmethod
    def teardown_class(cls):
        pass

    def setup_method(self, method):
        pass

    def teardown_method(self, method):
        pass

    def test__to_ast(self):
        pass

    def test__to_varname(self):
        pass



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.models.baseoperator import BaseOperator


class HiveOperator(BaseOperator):
    hql: "str"
    hive_cli_conn_id: "str"
    schema: "str"
    hiveconfs: "dict[Any, Any] | None"
    hiveconf_jinja_translate: "bool"
    script_begin_tag: "str | None"
    mapred_queue: "str | None"
    mapred_queue_priority: "str | None"
    mapred_job_name: "str | None"
    hive_cli_params: "str"
    auth: "str | None"
    proxy_user: "str | None"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.providers.google.cloud.operators.dataflow import (
    DataflowCreatePipelineOperator,
    DataflowRunPipelineOperator,
)


class CreateDataPipelineOperator(DataflowCreatePipelineOperator):
    pass


class RunDataPipelineOperator(DataflowRunPipelineOperator):
    pass



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.sensors.base import BaseSensorOperator


class DataformWorkflowInvocationStateSensor(BaseSensorOperator):
    project_id: "str"
    region: "str"
    repository_id: "str"
    workflow_invocation_id: "str"
    expected_statuses: "set[int] | int"
    failure_statuses: "Iterable[int] | None"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.sensors.base import BaseSensorOperator


class AzureCosmosDocumentSensor(BaseSensorOperator):
    database_name: "str"
    collection_name: "str"
    document_id: "str"
    azure_cosmos_conn_id: "str"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.providers.amazon.aws.sensors.base_aws import AwsBaseSensor


class EcsBaseSensor(AwsBaseSensor):
    aws_conn_id: "str | None"
    region_name: "str | None"
    verify: "bool | str | None"
    botocore_config: "dict | None"
    region: "str | None | ArgNotSet"


class EcsClusterStateSensor(EcsBaseSensor):
    cluster_name: "str"
    target_state: "EcsClusterStates | None"
    failure_states: "set[EcsClusterStates] | None"


class EcsTaskDefinitionStateSensor(EcsBaseSensor):
    task_definition: "str"
    target_state: "EcsTaskDefinitionStates | None"


class EcsTaskStateSensor(EcsBaseSensor):
    cluster: "str"
    task: "str"
    target_state: "EcsTaskStates | None"
    failure_states: "set[EcsTaskStates] | None"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.providers.amazon.aws.operators.base_aws import (
    AwsBaseOperator,
)


class SnsPublishOperator(AwsBaseOperator):
    target_arn: "str"
    message: "str"
    subject: "str | None"
    message_attributes: "dict | None"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.models.baseoperator import BaseOperator


class LocalFilesystemToGoogleDriveOperator(BaseOperator):
    local_paths: "Sequence[Path] | Sequence[str]"
    drive_folder: "Path | str"
    gcp_conn_id: "str"
    delete: "bool"
    ignore_if_missing: "bool"
    chunk_size: "int"
    resumable: "bool"
    delegate_to: "str | None"
    impersonation_chain: "str | Sequence[str] | None"
    folder_id: "str"
    show_full_target_path: "bool"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.models.baseoperator import BaseOperator


class TeradataToTeradataOperator(BaseOperator):
    dest_teradata_conn_id: "str"
    destination_table: "str"
    source_teradata_conn_id: "str"
    sql: "str"
    sql_params: "dict | None"
    rows_chunk: "int"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.sensors.base import BaseSensorOperator


class BigQueryDataTransferServiceTransferRunSensor(BaseSensorOperator):
    run_id: "str"
    transfer_config_id: "str"
    expected_statuses: "set[str | TransferState | int] | str | TransferState | int"
    project_id: "str"
    gcp_conn_id: "str"
    retry: "Retry | _MethodDefault"
    request_timeout: "float | None"
    metadata: "Sequence[tuple[str, str]]"
    location: "str | None"
    impersonation_chain: "str | Sequence[str] | None"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.sensors.base import BaseSensorOperator


class MetastoreHivePartitionSensor(BaseSensorOperator):
    service_id: "str"
    region: "str"
    table: "str"
    partitions: "list[str] | None"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.providers.google.cloud.operators.cloud_base import (
    GoogleCloudBaseOperator,
)


class DataprepGetJobsForJobGroupOperator(GoogleCloudBaseOperator):
    dataprep_conn_id: "str"
    job_group_id: "int | str"


class DataprepGetJobGroupOperator(GoogleCloudBaseOperator):
    dataprep_conn_id: "str"
    project_id: "str"
    job_group_id: "int | str"
    embed: "str"
    include_deleted: "bool"


class DataprepRunJobGroupOperator(GoogleCloudBaseOperator):
    project_id: "str"
    dataprep_conn_id: "str"
    body_request: "dict"


class DataprepCopyFlowOperator(GoogleCloudBaseOperator):
    project_id: "str"
    dataprep_conn_id: "str"
    flow_id: "int | str"
    name: "str"
    description: "str"
    copy_datasources: "bool"


class DataprepDeleteFlowOperator(GoogleCloudBaseOperator):
    dataprep_conn_id: "str"
    flow_id: "int | str"


class DataprepRunFlowOperator(GoogleCloudBaseOperator):
    project_id: "str"
    flow_id: "int | str"
    body_request: "dict"
    dataprep_conn_id: "str"



import pytest
from click.testing import CliRunner


@pytest.fixture()
def runner():
    return CliRunner()


def test_main(mocker, runner):
    from airfly.cli.main import main

    mock_tree = mocker.MagicMock()
    mock_TaskTree = mocker.patch("airfly.cli.main.TaskTree")

    mock_TaskTree.from_module.return_value = mock_tree

    result = runner.invoke(main, ["--modname", "not_existent_module"])
    assert result.exit_code == 1
    assert isinstance(result.exception, ModuleNotFoundError)

    dag_name = "my_dag"
    result = runner.invoke(main, ["--name", dag_name, "--modname", "airfly"])
    assert result.exit_code == 0
    assert mock_TaskTree.from_module.called
    assert mock_tree.to_dag.called
    assert dag_name in mock_tree.to_dag.call_args.args



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.providers.amazon.aws.sensors.base_aws import AwsBaseSensor


class AthenaSensor(AwsBaseSensor):
    query_execution_id: "str"
    max_retries: "int | None"
    sleep_time: "int"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.providers.google.cloud.operators.cloud_base import (
    GoogleCloudBaseOperator,
)


class PromptLanguageModelOperator(GoogleCloudBaseOperator):
    pass


class GenerateTextEmbeddingsOperator(GoogleCloudBaseOperator):
    pass


class PromptMultimodalModelOperator(GoogleCloudBaseOperator):
    pass


class PromptMultimodalModelWithMediaOperator(GoogleCloudBaseOperator):
    pass


class TextGenerationModelPredictOperator(GoogleCloudBaseOperator):
    project_id: "str"
    location: "str"
    prompt: "str"
    pretrained_model: "str"
    temperature: "float"
    max_output_tokens: "int"
    top_p: "float"
    top_k: "int"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class TextEmbeddingModelGetEmbeddingsOperator(GoogleCloudBaseOperator):
    project_id: "str"
    location: "str"
    prompt: "str"
    pretrained_model: "str"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class GenerativeModelGenerateContentOperator(GoogleCloudBaseOperator):
    project_id: "str"
    contents: "list"
    location: "str"
    tools: "list | None"
    generation_config: "dict | None"
    safety_settings: "dict | None"
    pretrained_model: "str"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.providers.google.cloud.operators.cloud_base import (
    GoogleCloudBaseOperator,
)


class LifeSciencesRunPipelineOperator(GoogleCloudBaseOperator):
    pass



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.providers.google.cloud.transfers.azure_blob_to_gcs import (
    AzureBlobStorageToGCSOperator,
)


class AzureBlobStorageToGCSOperator(AzureBlobStorageToGCSOperator):
    pass



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.models.baseoperator import BaseOperator


class JenkinsJobTriggerOperator(BaseOperator):
    jenkins_connection_id: "str"
    job_name: "str"
    parameters: "ParamType"
    sleep_time: "int"
    max_try_before_job_appears: "int"
    allowed_jenkins_states: "Iterable[str] | None"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.models.baseoperator import BaseOperator


class TriggerDagRunOperator(BaseOperator):
    trigger_dag_id: "str"
    trigger_run_id: "str | None"
    conf: "dict | None"
    logical_date: "str | datetime.datetime | None"
    reset_dag_run: "bool"
    wait_for_completion: "bool"
    poke_interval: "int"
    allowed_states: "list[str | DagRunState] | None"
    failed_states: "list[str | DagRunState] | None"
    skip_when_already_exists: "bool"
    deferrable: "bool"
    execution_date: "str | datetime.datetime | None"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.models.baseoperator import BaseOperator


class CloudFirestoreExportDatabaseOperator(BaseOperator):
    body: "dict"
    database_id: "str"
    project_id: "str"
    gcp_conn_id: "str"
    api_version: "str"
    impersonation_chain: "str | Sequence[str] | None"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.providers.google.cloud.operators.cloud_base import (
    GoogleCloudBaseOperator,
)


class CreateBatchPredictionJobOperator(GoogleCloudBaseOperator):
    region: "str"
    project_id: "str"
    job_display_name: "str"
    model_name: "str | Model"
    instances_format: "str"
    predictions_format: "str"
    gcs_source: "str | Sequence[str] | None"
    bigquery_source: "str | None"
    gcs_destination_prefix: "str | None"
    bigquery_destination_prefix: "str | None"
    model_parameters: "dict | None"
    machine_type: "str | None"
    accelerator_type: "str | None"
    accelerator_count: "int | None"
    starting_replica_count: "int | None"
    max_replica_count: "int | None"
    generate_explanation: "bool | None"
    explanation_metadata: "explain.ExplanationMetadata | None"
    explanation_parameters: "explain.ExplanationParameters | None"
    labels: "dict[str, str] | None"
    encryption_spec_key_name: "str | None"
    sync: "bool"
    create_request_timeout: "float | None"
    batch_size: "int | None"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"
    deferrable: "bool"
    poll_interval: "int"


class DeleteBatchPredictionJobOperator(GoogleCloudBaseOperator):
    region: "str"
    project_id: "str"
    batch_prediction_job_id: "str"
    retry: "Retry | _MethodDefault"
    timeout: "float | None"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class GetBatchPredictionJobOperator(GoogleCloudBaseOperator):
    region: "str"
    project_id: "str"
    batch_prediction_job: "str"
    retry: "Retry | _MethodDefault"
    timeout: "float | None"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class ListBatchPredictionJobsOperator(GoogleCloudBaseOperator):
    region: "str"
    project_id: "str"
    filter: "str | None"
    page_size: "int | None"
    page_token: "str | None"
    read_mask: "str | None"
    retry: "Retry | _MethodDefault"
    timeout: "float | None"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.models.baseoperator import BaseOperator


class GoogleCampaignManagerDeleteReportOperator(BaseOperator):
    profile_id: "str"
    report_name: "str | None"
    report_id: "str | None"
    api_version: "str"
    gcp_conn_id: "str"
    delegate_to: "str | None"
    impersonation_chain: "str | Sequence[str] | None"


class GoogleCampaignManagerDownloadReportOperator(BaseOperator):
    profile_id: "str"
    report_id: "str"
    file_id: "str"
    bucket_name: "str"
    report_name: "str | None"
    gzip: "bool"
    chunk_size: "int"
    api_version: "str"
    gcp_conn_id: "str"
    delegate_to: "str | None"
    impersonation_chain: "str | Sequence[str] | None"


class GoogleCampaignManagerInsertReportOperator(BaseOperator):
    profile_id: "str"
    report: "dict[str, Any]"
    api_version: "str"
    gcp_conn_id: "str"
    delegate_to: "str | None"
    impersonation_chain: "str | Sequence[str] | None"


class GoogleCampaignManagerRunReportOperator(BaseOperator):
    profile_id: "str"
    report_id: "str"
    synchronous: "bool"
    api_version: "str"
    gcp_conn_id: "str"
    delegate_to: "str | None"
    impersonation_chain: "str | Sequence[str] | None"


class GoogleCampaignManagerBatchInsertConversionsOperator(BaseOperator):
    profile_id: "str"
    conversions: "list[dict[str, Any]]"
    encryption_entity_type: "str"
    encryption_entity_id: "int"
    encryption_source: "str"
    max_failed_inserts: "int"
    api_version: "str"
    gcp_conn_id: "str"
    delegate_to: "str | None"
    impersonation_chain: "str | Sequence[str] | None"


class GoogleCampaignManagerBatchUpdateConversionsOperator(BaseOperator):
    profile_id: "str"
    conversions: "list[dict[str, Any]]"
    encryption_entity_type: "str"
    encryption_entity_id: "int"
    encryption_source: "str"
    max_failed_updates: "int"
    api_version: "str"
    gcp_conn_id: "str"
    delegate_to: "str | None"
    impersonation_chain: "str | Sequence[str] | None"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.models.baseoperator import BaseOperator


class RedshiftCreateClusterOperator(BaseOperator):
    cluster_identifier: "str"
    node_type: "str"
    master_username: "str"
    master_user_password: "str"
    cluster_type: "str"
    db_name: "str"
    number_of_nodes: "int"
    cluster_security_groups: "list[str] | None"
    vpc_security_group_ids: "list[str] | None"
    cluster_subnet_group_name: "str | None"
    availability_zone: "str | None"
    preferred_maintenance_window: "str | None"
    cluster_parameter_group_name: "str | None"
    automated_snapshot_retention_period: "int"
    manual_snapshot_retention_period: "int | None"
    port: "int"
    cluster_version: "str"
    allow_version_upgrade: "bool"
    publicly_accessible: "bool"
    encrypted: "bool"
    hsm_client_certificate_identifier: "str | None"
    hsm_configuration_identifier: "str | None"
    elastic_ip: "str | None"
    tags: "list[Any] | None"
    kms_key_id: "str | None"
    enhanced_vpc_routing: "bool"
    additional_info: "str | None"
    iam_roles: "list[str] | None"
    maintenance_track_name: "str | None"
    snapshot_schedule_identifier: "str | None"
    availability_zone_relocation: "bool | None"
    aqua_configuration_status: "str | None"
    default_iam_role_arn: "str | None"
    aws_conn_id: "str | None"
    wait_for_completion: "bool"
    max_attempt: "int"
    poll_interval: "int"
    deferrable: "bool"


class RedshiftCreateClusterSnapshotOperator(BaseOperator):
    snapshot_identifier: "str"
    cluster_identifier: "str"
    retention_period: "int"
    tags: "list[Any] | None"
    wait_for_completion: "bool"
    poll_interval: "int"
    max_attempt: "int"
    aws_conn_id: "str | None"
    deferrable: "bool"


class RedshiftDeleteClusterSnapshotOperator(BaseOperator):
    snapshot_identifier: "str"
    cluster_identifier: "str"
    wait_for_completion: "bool"
    aws_conn_id: "str | None"
    poll_interval: "int"


class RedshiftResumeClusterOperator(BaseOperator):
    cluster_identifier: "str"
    aws_conn_id: "str | None"
    wait_for_completion: "bool"
    deferrable: "bool"
    poll_interval: "int"
    max_attempts: "int"


class RedshiftPauseClusterOperator(BaseOperator):
    cluster_identifier: "str"
    aws_conn_id: "str | None"
    wait_for_completion: "bool"
    deferrable: "bool"
    poll_interval: "int"
    max_attempts: "int"


class RedshiftDeleteClusterOperator(BaseOperator):
    cluster_identifier: "str"
    skip_final_cluster_snapshot: "bool"
    final_cluster_snapshot_identifier: "str | None"
    wait_for_completion: "bool"
    aws_conn_id: "str | None"
    poll_interval: "int"
    deferrable: "bool"
    max_attempts: "int"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.sensors.base import BaseSensorOperator


class GithubSensor(BaseSensorOperator):
    method_name: "str"
    github_conn_id: "str"
    method_params: "dict | None"
    result_processor: "Callable | None"


class BaseGithubRepositorySensor(GithubSensor):
    github_conn_id: "str"
    repository_name: "str | None"
    result_processor: "Callable | None"


class GithubTagSensor(BaseGithubRepositorySensor):
    github_conn_id: "str"
    tag_name: "str | None"
    repository_name: "str | None"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.providers.google.cloud.operators.cloud_base import (
    GoogleCloudBaseOperator,
)


class CustomTrainingJobBaseOperator(GoogleCloudBaseOperator):
    project_id: "str"
    region: "str"
    display_name: "str"
    container_uri: "str"
    model_serving_container_image_uri: "str | None"
    model_serving_container_predict_route: "str | None"
    model_serving_container_health_route: "str | None"
    model_serving_container_command: "Sequence[str] | None"
    model_serving_container_args: "Sequence[str] | None"
    model_serving_container_environment_variables: "dict[str, str] | None"
    model_serving_container_ports: "Sequence[int] | None"
    model_description: "str | None"
    model_instance_schema_uri: "str | None"
    model_parameters_schema_uri: "str | None"
    model_prediction_schema_uri: "str | None"
    parent_model: "str | None"
    is_default_version: "bool | None"
    model_version_aliases: "list[str] | None"
    model_version_description: "str | None"
    labels: "dict[str, str] | None"
    training_encryption_spec_key_name: "str | None"
    model_encryption_spec_key_name: "str | None"
    staging_bucket: "str | None"
    dataset_id: "str | None"
    annotation_schema_uri: "str | None"
    model_display_name: "str | None"
    model_labels: "dict[str, str] | None"
    base_output_dir: "str | None"
    service_account: "str | None"
    network: "str | None"
    bigquery_destination: "str | None"
    args: "list[str | float | int] | None"
    environment_variables: "dict[str, str] | None"
    replica_count: "int"
    machine_type: "str"
    accelerator_type: "str"
    accelerator_count: "int"
    boot_disk_type: "str"
    boot_disk_size_gb: "int"
    training_fraction_split: "float | None"
    validation_fraction_split: "float | None"
    test_fraction_split: "float | None"
    training_filter_split: "str | None"
    validation_filter_split: "str | None"
    test_filter_split: "str | None"
    predefined_split_column_name: "str | None"
    timestamp_split_column_name: "str | None"
    tensorboard: "str | None"
    sync: "_empty"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class CreateCustomContainerTrainingJobOperator(CustomTrainingJobBaseOperator):
    command: "Sequence[str]"
    region: "str"
    parent_model: "str | None"
    impersonation_chain: "str | Sequence[str] | None"
    dataset_id: "str | None"
    deferrable: "bool"
    poll_interval: "int"


class CreateCustomPythonPackageTrainingJobOperator(CustomTrainingJobBaseOperator):
    python_package_gcs_uri: "str"
    python_module_name: "str"
    region: "str"
    parent_model: "str | None"
    impersonation_chain: "str | Sequence[str] | None"
    dataset_id: "str | None"
    deferrable: "bool"
    poll_interval: "int"


class CreateCustomTrainingJobOperator(CustomTrainingJobBaseOperator):
    script_path: "str"
    requirements: "Sequence[str] | None"
    region: "str"
    parent_model: "str | None"
    impersonation_chain: "str | Sequence[str] | None"
    dataset_id: "str | None"
    deferrable: "bool"
    poll_interval: "int"


class DeleteCustomTrainingJobOperator(GoogleCloudBaseOperator):
    training_pipeline_id: "str"
    custom_job_id: "str"
    region: "str"
    project_id: "str"
    retry: "Retry | _MethodDefault"
    timeout: "float | None"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class ListCustomTrainingJobOperator(GoogleCloudBaseOperator):
    region: "str"
    project_id: "str"
    page_size: "int | None"
    page_token: "str | None"
    filter: "str | None"
    read_mask: "str | None"
    retry: "Retry | _MethodDefault"
    timeout: "float | None"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.providers.google.cloud.operators.cloud_base import (
    GoogleCloudBaseOperator,
)


class DataformCreateCompilationResultOperator(GoogleCloudBaseOperator):
    project_id: "str"
    region: "str"
    repository_id: "str"
    compilation_result: "CompilationResult | dict"
    retry: "Retry | _MethodDefault"
    timeout: "float | None"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class DataformGetCompilationResultOperator(GoogleCloudBaseOperator):
    project_id: "str"
    region: "str"
    repository_id: "str"
    compilation_result_id: "str"
    retry: "Retry | _MethodDefault"
    timeout: "float | None"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class DataformCreateWorkflowInvocationOperator(GoogleCloudBaseOperator):
    project_id: "str"
    region: "str"
    repository_id: "str"
    workflow_invocation: "WorkflowInvocation | dict"
    retry: "Retry | _MethodDefault"
    timeout: "int | None"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"
    asynchronous: "bool"
    wait_time: "int"


class DataformGetWorkflowInvocationOperator(GoogleCloudBaseOperator):
    project_id: "str"
    region: "str"
    repository_id: "str"
    workflow_invocation_id: "str"
    retry: "Retry | _MethodDefault"
    timeout: "float | None"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class DataformQueryWorkflowInvocationActionsOperator(GoogleCloudBaseOperator):
    project_id: "str"
    region: "str"
    repository_id: "str"
    workflow_invocation_id: "str"
    retry: "Retry | _MethodDefault"
    timeout: "float | None"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class DataformCancelWorkflowInvocationOperator(GoogleCloudBaseOperator):
    project_id: "str"
    region: "str"
    repository_id: "str"
    workflow_invocation_id: "str"
    retry: "Retry | _MethodDefault"
    timeout: "float | None"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class DataformCreateRepositoryOperator(GoogleCloudBaseOperator):
    project_id: "str"
    region: "str"
    repository_id: "str"
    retry: "Retry | _MethodDefault"
    timeout: "float | None"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class DataformDeleteRepositoryOperator(GoogleCloudBaseOperator):
    project_id: "str"
    region: "str"
    repository_id: "str"
    force: "bool"
    retry: "Retry | _MethodDefault"
    timeout: "float | None"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class DataformCreateWorkspaceOperator(GoogleCloudBaseOperator):
    project_id: "str"
    region: "str"
    repository_id: "str"
    workspace_id: "str"
    retry: "Retry | _MethodDefault"
    timeout: "float | None"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class DataformDeleteWorkspaceOperator(GoogleCloudBaseOperator):
    project_id: "str"
    region: "str"
    repository_id: "str"
    workspace_id: "str"
    retry: "Retry | _MethodDefault"
    timeout: "float | None"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class DataformWriteFileOperator(GoogleCloudBaseOperator):
    project_id: "str"
    region: "str"
    repository_id: "str"
    workspace_id: "str"
    filepath: "str"
    contents: "bytes"
    retry: "Retry | _MethodDefault"
    timeout: "float | None"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class DataformMakeDirectoryOperator(GoogleCloudBaseOperator):
    project_id: "str"
    region: "str"
    repository_id: "str"
    workspace_id: "str"
    directory_path: "str"
    retry: "Retry | _MethodDefault"
    timeout: "float | None"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class DataformRemoveFileOperator(GoogleCloudBaseOperator):
    project_id: "str"
    region: "str"
    repository_id: "str"
    workspace_id: "str"
    filepath: "str"
    retry: "Retry | _MethodDefault"
    timeout: "float | None"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class DataformRemoveDirectoryOperator(GoogleCloudBaseOperator):
    project_id: "str"
    region: "str"
    repository_id: "str"
    workspace_id: "str"
    directory_path: "str"
    retry: "Retry | _MethodDefault"
    timeout: "float | None"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class DataformInstallNpmPackagesOperator(GoogleCloudBaseOperator):
    project_id: "str"
    region: "str"
    repository_id: "str"
    workspace_id: "str"
    retry: "Retry | _MethodDefault"
    timeout: "float | None"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.models.baseoperator import BaseOperator


class SSHOperator(BaseOperator):
    ssh_hook: "SSHHook | None"
    ssh_conn_id: "str | None"
    remote_host: "str | None"
    command: "str | None"
    conn_timeout: "int | None"
    cmd_timeout: "int | ArgNotSet | None"
    environment: "dict | None"
    get_pty: "bool"
    banner_timeout: "float"
    skip_on_exit_code: "int | Container[int] | None"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.models.baseoperator import BaseOperator


class SparkSqlOperator(BaseOperator):
    sql: "str"
    conf: "str | None"
    conn_id: "str"
    total_executor_cores: "int | None"
    executor_cores: "int | None"
    executor_memory: "str | None"
    keytab: "str | None"
    principal: "str | None"
    master: "str | None"
    name: "str"
    num_executors: "int | None"
    verbose: "bool"
    yarn_queue: "str | None"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.models.baseoperator import BaseOperator


class AnalyticDBSparkBaseOperator(BaseOperator):
    adb_spark_conn_id: "str"
    region: "str | None"
    polling_interval: "int"


class AnalyticDBSparkSQLOperator(AnalyticDBSparkBaseOperator):
    sql: "str"
    conf: "dict[Any, Any] | None"
    driver_resource_spec: "str | None"
    executor_resource_spec: "str | None"
    num_executors: "int | str | None"
    name: "str | None"
    cluster_id: "str"
    rg_name: "str"


class AnalyticDBSparkBatchOperator(AnalyticDBSparkBaseOperator):
    file: "str"
    class_name: "str | None"
    args: "Sequence[str | int | float] | None"
    conf: "dict[Any, Any] | None"
    jars: "Sequence[str] | None"
    py_files: "Sequence[str] | None"
    files: "Sequence[str] | None"
    driver_resource_spec: "str | None"
    executor_resource_spec: "str | None"
    num_executors: "int | str | None"
    archives: "Sequence[str] | None"
    name: "str | None"
    cluster_id: "str"
    rg_name: "str"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.models.baseoperator import BaseOperator


class S3ToFTPOperator(BaseOperator):
    s3_bucket: "_empty"
    s3_key: "_empty"
    ftp_path: "_empty"
    aws_conn_id: "_empty"
    ftp_conn_id: "_empty"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.sensors.base import BaseSensorOperator


class NamedHivePartitionSensor(BaseSensorOperator):
    partition_names: "list[str]"
    metastore_conn_id: "str"
    poke_interval: "int"
    hook: "Any"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.models.baseoperator import BaseOperator


class TrinoToMySqlOperator(BaseOperator):
    sql: "str"
    mysql_table: "str"
    trino_conn_id: "str"
    mysql_conn_id: "str"
    mysql_preoperator: "str | None"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.providers.google.cloud.operators.cloud_base import (
    GoogleCloudBaseOperator,
)


class RunPipelineJobOperator(GoogleCloudBaseOperator):
    project_id: "str"
    region: "str"
    display_name: "str"
    template_path: "str"
    job_id: "str | None"
    pipeline_root: "str | None"
    parameter_values: "dict[str, Any] | None"
    input_artifacts: "dict[str, str] | None"
    enable_caching: "bool | None"
    encryption_spec_key_name: "str | None"
    labels: "dict[str, str] | None"
    failure_policy: "str | None"
    service_account: "str | None"
    network: "str | None"
    create_request_timeout: "float | None"
    experiment: "str | experiment_resources.Experiment | None"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"
    deferrable: "bool"
    poll_interval: "int"


class GetPipelineJobOperator(GoogleCloudBaseOperator):
    project_id: "str"
    region: "str"
    pipeline_job_id: "str"
    retry: "Retry | _MethodDefault"
    timeout: "float | None"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class ListPipelineJobOperator(GoogleCloudBaseOperator):
    region: "str"
    project_id: "str"
    page_size: "int | None"
    page_token: "str | None"
    filter: "str | None"
    order_by: "str | None"
    retry: "Retry | _MethodDefault"
    timeout: "float | None"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class DeletePipelineJobOperator(GoogleCloudBaseOperator):
    project_id: "str"
    region: "str"
    pipeline_job_id: "str"
    retry: "Retry | _MethodDefault"
    timeout: "float | None"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.providers.google.cloud.operators.cloud_base import (
    GoogleCloudBaseOperator,
)


class WorkflowsCreateWorkflowOperator(GoogleCloudBaseOperator):
    workflow: "dict"
    workflow_id: "str"
    location: "str"
    project_id: "str"
    retry: "Retry | _MethodDefault"
    timeout: "float | None"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    force_rerun: "bool"
    impersonation_chain: "str | Sequence[str] | None"


class WorkflowsUpdateWorkflowOperator(GoogleCloudBaseOperator):
    workflow_id: "str"
    location: "str"
    project_id: "str"
    update_mask: "FieldMask | None"
    retry: "Retry | _MethodDefault"
    timeout: "float | None"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class WorkflowsDeleteWorkflowOperator(GoogleCloudBaseOperator):
    workflow_id: "str"
    location: "str"
    project_id: "str"
    retry: "Retry | _MethodDefault"
    timeout: "float | None"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class WorkflowsListWorkflowsOperator(GoogleCloudBaseOperator):
    location: "str"
    project_id: "str"
    filter_: "str | None"
    order_by: "str | None"
    retry: "Retry | _MethodDefault"
    timeout: "float | None"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class WorkflowsGetWorkflowOperator(GoogleCloudBaseOperator):
    workflow_id: "str"
    location: "str"
    project_id: "str"
    retry: "Retry | _MethodDefault"
    timeout: "float | None"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class WorkflowsCreateExecutionOperator(GoogleCloudBaseOperator):
    workflow_id: "str"
    execution: "dict"
    location: "str"
    project_id: "str"
    retry: "Retry | _MethodDefault"
    timeout: "float | None"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class WorkflowsCancelExecutionOperator(GoogleCloudBaseOperator):
    workflow_id: "str"
    execution_id: "str"
    location: "str"
    project_id: "str"
    retry: "Retry | _MethodDefault"
    timeout: "float | None"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class WorkflowsListExecutionsOperator(GoogleCloudBaseOperator):
    workflow_id: "str"
    location: "str"
    start_date_filter: "datetime.datetime | None"
    project_id: "str"
    retry: "Retry | _MethodDefault"
    timeout: "float | None"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class WorkflowsGetExecutionOperator(GoogleCloudBaseOperator):
    workflow_id: "str"
    execution_id: "str"
    location: "str"
    project_id: "str"
    retry: "Retry | _MethodDefault"
    timeout: "float | None"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.providers.amazon.aws.sensors.base_aws import AwsBaseSensor


class StepFunctionExecutionSensor(AwsBaseSensor):
    execution_arn: "str"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.sensors.base import BaseSensorOperator


class AirbyteJobSensor(BaseSensorOperator):
    airbyte_job_id: "int"
    deferrable: "bool"
    airbyte_conn_id: "str"
    api_version: "str"
    api_type: "Literal['config', 'cloud']"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.models.baseoperator import BaseOperator


class WeaviateIngestOperator(BaseOperator):
    conn_id: "str"
    collection_name: "str"
    input_data: "list[dict[str, Any]] | pd.DataFrame | None"
    vector_col: "str"
    uuid_column: "str"
    tenant: "str | None"
    hook_params: "dict | None"
    input_json: "list[dict[str, Any]] | pd.DataFrame | None"


class WeaviateDocumentIngestOperator(BaseOperator):
    conn_id: "str"
    input_data: "pd.DataFrame | list[dict[str, Any]] | list[pd.DataFrame]"
    collection_name: "str"
    document_column: "str"
    existing: "str"
    uuid_column: "str"
    vector_col: "str"
    tenant: "str | None"
    verbose: "bool"
    hook_params: "dict | None"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.models.baseoperator import BaseOperator


class SalesforceToGcsOperator(BaseOperator):
    query: "str"
    bucket_name: "str"
    object_name: "str"
    salesforce_conn_id: "str"
    include_deleted: "bool"
    query_params: "dict | None"
    export_format: "str"
    coerce_to_timestamp: "bool"
    record_time_added: "bool"
    gzip: "bool"
    gcp_conn_id: "str"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.operators.empty import EmptyOperator
from airfly._vendor.airflow.sensors.base import BaseSensorOperator


class ExternalTaskMarker(EmptyOperator):
    external_dag_id: "str"
    external_task_id: "str"
    execution_date: "str | datetime.datetime | None"
    recursion_depth: "int"


class ExternalTaskSensor(BaseSensorOperator):
    external_dag_id: "str"
    external_task_id: "str | None"
    external_task_ids: "Collection[str] | None"
    external_task_group_id: "str | None"
    allowed_states: "Iterable[str] | None"
    skipped_states: "Iterable[str] | None"
    failed_states: "Iterable[str] | None"
    execution_delta: "datetime.timedelta | None"
    execution_date_fn: "Callable | None"
    check_existence: "bool"
    poll_interval: "float"
    deferrable: "bool"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.providers.amazon.aws.sensors.base_aws import AwsBaseSensor


class DynamoDBValueSensor(AwsBaseSensor):
    table_name: "str"
    partition_key_name: "str"
    partition_key_value: "str"
    attribute_name: "str"
    attribute_value: "str | Iterable[str]"
    sort_key_name: "str | None"
    sort_key_value: "str | None"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.models.baseoperator import BaseOperator


class SlackAPIOperator(BaseOperator):
    slack_conn_id: "str"
    method: "str | None"
    api_params: "dict | None"
    base_url: "str | None"
    proxy: "str | None"
    timeout: "int | None"
    retry_handlers: "list[RetryHandler] | None"


class SlackAPIPostOperator(SlackAPIOperator):
    channel: "str"
    username: "str"
    text: "str"
    icon_url: "str"
    blocks: "list | None"
    attachments: "list | None"


class SlackAPIFileOperator(SlackAPIOperator):
    channels: "str | Sequence[str] | None"
    initial_comment: "str | None"
    filename: "str | None"
    filetype: "str | None"
    content: "str | None"
    title: "str | None"
    method_version: "Literal['v1', 'v2']"
    channel: "str | Sequence[str] | None | ArgNotSet"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.models.baseoperator import BaseOperator
from airfly._vendor.airflow.models.skipmixin import SkipMixin


class BaseSensorOperator(BaseOperator, SkipMixin):
    poke_interval: "timedelta | float"
    timeout: "timedelta | float"
    soft_fail: "bool"
    mode: "str"
    exponential_backoff: "bool"
    max_wait: "timedelta | float | None"
    silent_fail: "bool"
    never_fail: "bool"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.models.baseoperator import BaseOperator
from airfly._vendor.airflow.models.skipmixin import SkipMixin


class BaseSQLOperator(BaseOperator):
    conn_id: "str | None"
    database: "str | None"
    hook_params: "dict | None"
    retry_on_failure: "bool"


class SQLExecuteQueryOperator(BaseSQLOperator):
    sql: "str | list[str]"
    autocommit: "bool"
    parameters: "Mapping | Iterable | None"
    handler: "Callable[[Any], Any]"
    conn_id: "str | None"
    database: "str | None"
    split_statements: "bool | None"
    return_last: "bool"
    show_return_value_in_logs: "bool"


class SQLCheckOperator(BaseSQLOperator):
    sql: "str"
    conn_id: "str | None"
    database: "str | None"
    parameters: "Iterable | Mapping[str, Any] | None"


class SQLColumnCheckOperator(BaseSQLOperator):
    table: "str"
    column_mapping: "dict[str, dict[str, Any]]"
    partition_clause: "str | None"
    conn_id: "str | None"
    database: "str | None"
    accept_none: "bool"


class SQLTableCheckOperator(BaseSQLOperator):
    table: "str"
    checks: "dict[str, dict[str, Any]]"
    partition_clause: "str | None"
    conn_id: "str | None"
    database: "str | None"


class SQLValueCheckOperator(BaseSQLOperator):
    sql: "str"
    pass_value: "Any"
    tolerance: "Any"
    conn_id: "str | None"
    database: "str | None"


class SQLIntervalCheckOperator(BaseSQLOperator):
    table: "str"
    metrics_thresholds: "dict[str, int]"
    date_filter_column: "str | None"
    days_back: "SupportsAbs[int]"
    ratio_formula: "str | None"
    ignore_zero: "bool"
    conn_id: "str | None"
    database: "str | None"


class SQLThresholdCheckOperator(BaseSQLOperator):
    sql: "str"
    min_threshold: "Any"
    max_threshold: "Any"
    conn_id: "str | None"
    database: "str | None"


class BranchSQLOperator(BaseSQLOperator, SkipMixin):
    sql: "str"
    follow_task_ids_if_true: "list[str]"
    follow_task_ids_if_false: "list[str]"
    conn_id: "str"
    database: "str | None"
    parameters: "Iterable | Mapping[str, Any] | None"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.models.baseoperator import BaseOperator


class MongoToS3Operator(BaseOperator):
    mongo_conn_id: "str"
    aws_conn_id: "str | None"
    mongo_collection: "str"
    mongo_query: "list | dict"
    s3_bucket: "str"
    s3_key: "str"
    mongo_db: "str | None"
    mongo_projection: "list | dict | None"
    replace: "bool"
    allow_disk_use: "bool"
    compression: "str | None"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.models.baseoperator import BaseOperator


class ExasolToS3Operator(BaseOperator):
    query_or_table: "str"
    key: "str"
    bucket_name: "str | None"
    replace: "bool"
    encrypt: "bool"
    gzip: "bool"
    acl_policy: "str | None"
    query_params: "dict | None"
    export_params: "dict | None"
    exasol_conn_id: "str"
    aws_conn_id: "str | None"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.abc import ABC
from airfly._vendor.airflow.models.baseoperator import BaseOperator


class DatabricksCreateJobsOperator(BaseOperator):
    json: "Any | None"
    name: "str | None"
    description: "str | None"
    tags: "dict[str, str] | None"
    tasks: "list[dict] | None"
    job_clusters: "list[dict] | None"
    email_notifications: "dict | None"
    webhook_notifications: "dict | None"
    notification_settings: "dict | None"
    timeout_seconds: "int | None"
    schedule: "dict | None"
    max_concurrent_runs: "int | None"
    git_source: "dict | None"
    access_control_list: "list[dict] | None"
    databricks_conn_id: "str"
    polling_period_seconds: "int"
    databricks_retry_limit: "int"
    databricks_retry_delay: "int"
    databricks_retry_args: "dict[Any, Any] | None"


class DatabricksSubmitRunOperator(BaseOperator):
    json: "Any | None"
    tasks: "list[object] | None"
    spark_jar_task: "dict[str, str] | None"
    notebook_task: "dict[str, str] | None"
    spark_python_task: "dict[str, str | list[str]] | None"
    spark_submit_task: "dict[str, list[str]] | None"
    pipeline_task: "dict[str, str] | None"
    dbt_task: "dict[str, str | list[str]] | None"
    new_cluster: "dict[str, object] | None"
    existing_cluster_id: "str | None"
    libraries: "list[dict[str, Any]] | None"
    run_name: "str | None"
    timeout_seconds: "int | None"
    databricks_conn_id: "str"
    polling_period_seconds: "int"
    databricks_retry_limit: "int"
    databricks_retry_delay: "int"
    databricks_retry_args: "dict[Any, Any] | None"
    do_xcom_push: "bool"
    idempotency_token: "str | None"
    access_control_list: "list[dict[str, str]] | None"
    wait_for_termination: "bool"
    git_source: "dict[str, str] | None"
    deferrable: "bool"


class DatabricksSubmitRunDeferrableOperator(DatabricksSubmitRunOperator):
    pass


class DatabricksRunNowOperator(BaseOperator):
    job_id: "str | None"
    job_name: "str | None"
    json: "Any | None"
    notebook_params: "dict[str, str] | None"
    python_params: "list[str] | None"
    jar_params: "list[str] | None"
    spark_submit_params: "list[str] | None"
    python_named_params: "dict[str, str] | None"
    idempotency_token: "str | None"
    databricks_conn_id: "str"
    polling_period_seconds: "int"
    databricks_retry_limit: "int"
    databricks_retry_delay: "int"
    databricks_retry_args: "dict[Any, Any] | None"
    do_xcom_push: "bool"
    wait_for_termination: "bool"
    deferrable: "bool"
    repair_run: "bool"
    cancel_previous_runs: "bool"


class DatabricksRunNowDeferrableOperator(DatabricksRunNowOperator):
    pass


class DatabricksTaskBaseOperator(BaseOperator, ABC):
    caller: "str"
    databricks_conn_id: "str"
    databricks_retry_args: "dict[Any, Any] | None"
    databricks_retry_delay: "int"
    databricks_retry_limit: "int"
    deferrable: "bool"
    existing_cluster_id: "str"
    job_cluster_key: "str"
    new_cluster: "dict[str, Any] | None"
    polling_period_seconds: "int"
    wait_for_termination: "bool"
    workflow_run_metadata: "dict[str, Any] | None"


class DatabricksNotebookOperator(DatabricksTaskBaseOperator):
    notebook_path: "str"
    source: "str"
    databricks_conn_id: "str"
    databricks_retry_args: "dict[Any, Any] | None"
    databricks_retry_delay: "int"
    databricks_retry_limit: "int"
    deferrable: "bool"
    existing_cluster_id: "str"
    job_cluster_key: "str"
    new_cluster: "dict[str, Any] | None"
    notebook_packages: "list[dict[str, Any]] | None"
    notebook_params: "dict | None"
    polling_period_seconds: "int"
    wait_for_termination: "bool"
    workflow_run_metadata: "dict | None"


class DatabricksTaskOperator(DatabricksTaskBaseOperator):
    task_config: "dict"
    databricks_conn_id: "str"
    databricks_retry_args: "dict[Any, Any] | None"
    databricks_retry_delay: "int"
    databricks_retry_limit: "int"
    deferrable: "bool"
    existing_cluster_id: "str"
    job_cluster_key: "str"
    new_cluster: "dict[str, Any] | None"
    polling_period_seconds: "int"
    wait_for_termination: "bool"
    workflow_run_metadata: "dict | None"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.providers.google.cloud.operators.cloud_base import (
    GoogleCloudBaseOperator,
)


class GCSCreateBucketOperator(GoogleCloudBaseOperator):
    bucket_name: "str"
    resource: "dict | None"
    storage_class: "str"
    location: "str"
    project_id: "str"
    labels: "dict | None"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class GCSDeleteBucketOperator(GoogleCloudBaseOperator):
    bucket_name: "str"
    force: "bool"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"
    user_project: "str | None"


class GCSListObjectsOperator(GoogleCloudBaseOperator):
    bucket: "str"
    prefix: "str | list[str] | None"
    delimiter: "str | None"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"
    match_glob: "str | None"


class GCSDeleteObjectsOperator(GoogleCloudBaseOperator):
    bucket_name: "str"
    objects: "list[str] | None"
    prefix: "str | list[str] | None"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class GCSBucketCreateAclEntryOperator(GoogleCloudBaseOperator):
    bucket: "str"
    entity: "str"
    role: "str"
    user_project: "str | None"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class GCSObjectCreateAclEntryOperator(GoogleCloudBaseOperator):
    bucket: "str"
    object_name: "str"
    entity: "str"
    role: "str"
    generation: "int | None"
    user_project: "str | None"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class GCSFileTransformOperator(GoogleCloudBaseOperator):
    source_bucket: "str"
    source_object: "str"
    transform_script: "str | list[str]"
    destination_bucket: "str | None"
    destination_object: "str | None"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class GCSTimeSpanFileTransformOperator(GoogleCloudBaseOperator):
    source_bucket: "str"
    source_prefix: "str"
    source_gcp_conn_id: "str"
    destination_bucket: "str"
    destination_prefix: "str"
    destination_gcp_conn_id: "str"
    transform_script: "str | list[str]"
    source_impersonation_chain: "str | Sequence[str] | None"
    destination_impersonation_chain: "str | Sequence[str] | None"
    chunk_size: "int | None"
    download_continue_on_fail: "bool | None"
    download_num_attempts: "int"
    upload_continue_on_fail: "bool | None"
    upload_num_attempts: "int"


class GCSSynchronizeBucketsOperator(GoogleCloudBaseOperator):
    source_bucket: "str"
    destination_bucket: "str"
    source_object: "str | None"
    destination_object: "str | None"
    recursive: "bool"
    delete_extra_files: "bool"
    allow_overwrite: "bool"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.models.baseoperator import BaseOperator


class AzureContainerInstancesOperator(BaseOperator):
    ci_conn_id: "str"
    resource_group: "str"
    name: "str"
    image: "str"
    region: "str"
    registry_conn_id: "str | None"
    environment_variables: "dict | None"
    secured_variables: "str | None"
    volumes: "list | None"
    memory_in_gb: "Any | None"
    cpu: "Any | None"
    gpu: "Any | None"
    command: "list[str] | None"
    remove_on_error: "bool"
    fail_if_exists: "bool"
    tags: "dict[str, str] | None"
    os_type: "str"
    restart_policy: "str"
    ip_address: "IpAddress | None"
    ports: "list[ContainerPort] | None"
    subnet_ids: "list[ContainerGroupSubnetId] | None"
    dns_config: "DnsConfiguration | None"
    diagnostics: "ContainerGroupDiagnostics | None"
    priority: "str | None"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.providers.amazon.aws.operators.s3 import S3ListOperator


class S3ToGCSOperator(S3ListOperator):
    bucket: "_empty"
    prefix: "_empty"
    apply_gcs_prefix: "_empty"
    delimiter: "_empty"
    aws_conn_id: "_empty"
    verify: "_empty"
    gcp_conn_id: "_empty"
    dest_gcs: "_empty"
    replace: "_empty"
    gzip: "_empty"
    google_impersonation_chain: "str | Sequence[str] | None"
    deferrable: "_empty"
    poll_interval: "int"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.models.baseoperator import BaseOperator


class GlacierToGCSOperator(BaseOperator):
    aws_conn_id: "str | None"
    gcp_conn_id: "str"
    vault_name: "str"
    bucket_name: "str"
    object_name: "str"
    gzip: "bool"
    chunk_size: "int"
    google_impersonation_chain: "str | Sequence[str] | None"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.providers.google.cloud.operators.cloud_base import (
    GoogleCloudBaseOperator,
)


class AutoMLTrainModelOperator(GoogleCloudBaseOperator):
    model: "dict"
    location: "str"
    project_id: "str"
    metadata: "MetaData"
    timeout: "float | None"
    retry: "Retry | _MethodDefault"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class AutoMLPredictOperator(GoogleCloudBaseOperator):
    model_id: "str | None"
    endpoint_id: "str | None"
    location: "str"
    payload: "dict"
    operation_params: "dict[str, str] | None"
    instances: "list[str] | None"
    project_id: "str"
    metadata: "MetaData"
    timeout: "float | None"
    retry: "Retry | _MethodDefault"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class AutoMLBatchPredictOperator(GoogleCloudBaseOperator):
    model_id: "str"
    input_config: "dict"
    output_config: "dict"
    location: "str"
    project_id: "str"
    prediction_params: "dict[str, str] | None"
    metadata: "MetaData"
    timeout: "float | None"
    retry: "Retry | _MethodDefault"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class AutoMLCreateDatasetOperator(GoogleCloudBaseOperator):
    dataset: "dict"
    location: "str"
    project_id: "str"
    metadata: "MetaData"
    timeout: "float | None"
    retry: "Retry | _MethodDefault"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class AutoMLImportDataOperator(GoogleCloudBaseOperator):
    dataset_id: "str"
    location: "str"
    input_config: "dict"
    project_id: "str"
    metadata: "MetaData"
    timeout: "float | None"
    retry: "Retry | _MethodDefault"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class AutoMLTablesListColumnSpecsOperator(GoogleCloudBaseOperator):
    dataset_id: "str"
    table_spec_id: "str"
    location: "str"
    field_mask: "dict | None"
    filter_: "str | None"
    page_size: "int | None"
    project_id: "str"
    metadata: "MetaData"
    timeout: "float | None"
    retry: "Retry | _MethodDefault"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class AutoMLTablesUpdateDatasetOperator(GoogleCloudBaseOperator):
    dataset: "dict"
    location: "str"
    update_mask: "dict | None"
    metadata: "MetaData"
    timeout: "float | None"
    retry: "Retry | _MethodDefault"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class AutoMLGetModelOperator(GoogleCloudBaseOperator):
    model_id: "str"
    location: "str"
    project_id: "str"
    metadata: "MetaData"
    timeout: "float | None"
    retry: "Retry | _MethodDefault"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class AutoMLDeleteModelOperator(GoogleCloudBaseOperator):
    model_id: "str"
    location: "str"
    project_id: "str"
    metadata: "MetaData"
    timeout: "float | None"
    retry: "Retry | _MethodDefault"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class AutoMLDeployModelOperator(GoogleCloudBaseOperator):
    model_id: "str"
    location: "str"
    project_id: "str"
    image_detection_metadata: "dict | None"
    metadata: "Sequence[tuple[str, str]]"
    timeout: "float | None"
    retry: "Retry | _MethodDefault"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class AutoMLTablesListTableSpecsOperator(GoogleCloudBaseOperator):
    dataset_id: "str"
    location: "str"
    page_size: "int | None"
    filter_: "str | None"
    project_id: "str"
    metadata: "MetaData"
    timeout: "float | None"
    retry: "Retry | _MethodDefault"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class AutoMLListDatasetOperator(GoogleCloudBaseOperator):
    location: "str"
    project_id: "str"
    metadata: "MetaData"
    timeout: "float | None"
    retry: "Retry | _MethodDefault"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class AutoMLDeleteDatasetOperator(GoogleCloudBaseOperator):
    dataset_id: "str | list[str]"
    location: "str"
    project_id: "str"
    metadata: "MetaData"
    timeout: "float | None"
    retry: "Retry | _MethodDefault"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.providers.google.cloud.operators.cloud_base import (
    GoogleCloudBaseOperator,
)


class CloudVideoIntelligenceDetectVideoLabelsOperator(GoogleCloudBaseOperator):
    input_uri: "str"
    input_content: "bytes | None"
    output_uri: "str | None"
    video_context: "dict | VideoContext | None"
    location: "str | None"
    retry: "Retry | _MethodDefault"
    timeout: "float | None"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class CloudVideoIntelligenceDetectVideoExplicitContentOperator(GoogleCloudBaseOperator):
    input_uri: "str"
    output_uri: "str | None"
    input_content: "bytes | None"
    video_context: "dict | VideoContext | None"
    location: "str | None"
    retry: "Retry | _MethodDefault"
    timeout: "float | None"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class CloudVideoIntelligenceDetectVideoShotsOperator(GoogleCloudBaseOperator):
    input_uri: "str"
    output_uri: "str | None"
    input_content: "bytes | None"
    video_context: "dict | VideoContext | None"
    location: "str | None"
    retry: "Retry | _MethodDefault"
    timeout: "float | None"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"



import inspect
import os
from collections import deque
from functools import lru_cache
from types import FunctionType, MethodType, ModuleType
from typing import (
    Any,
    Dict,
    Generator,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    Union,
)

import asttrs
import attr
import loguru
import networkx as nx
import regex as re

import airfly
from airfly._vendor import collect_airflow_operators
from airfly.utils import (
    blacking,
    collect_objects,
    immutable,
    isorting,
    issubclass_by_qualname,
    qualname,
)

TaskClass = Type["Task"]


BUILTIN_OPERATORS = collect_airflow_operators()


class Literal:

    def __init__(self, expr: str, deps: List[Any] = None):
        self.expr = expr
        self.deps = deps if isinstance(deps, List) else [deps]

    def __repr__(self):
        return self.expr

    def _to_ast(self):
        return asttrs.Name(id=self.expr, ctx=asttrs.Load())

    def _dep_ast(self, param_ctx: "ParamContext" = None) -> List[asttrs.stmt]:
        body = []

        for d in self.deps:

            if isinstance(d, Literal):
                body.append(d._to_ast())  # NOTE: alias conflict may occur
            else:
                body.extend(ParamContext.get(d, param_ctx)._dep_ast(param_ctx))

        return body


class Param:
    def __init__(self, target: Any, alias: str = None):
        self.target = target
        self.alias = alias or (
            target.__qualname__.split(".")[0]
            if hasattr(target, "__qualname__")
            else None
        )
        self.conflicts = 0

    def _target_ast(self, param_ctx: "ParamContext" = None) -> asttrs.stmt:
        value = self.target
        if isinstance(value, (type(None), bool, str, int, float)):  # early return
            return asttrs.Constant(value=value)

        if isinstance(value, List):
            return asttrs.List(
                elts=[
                    ParamContext.get(el, param_ctx)._target_ast(param_ctx)
                    for el in value
                ],
                ctx=asttrs.Load(),
            )

        if isinstance(value, Tuple):
            return asttrs.Tuple(
                elts=[
                    ParamContext.get(el, param_ctx)._target_ast(param_ctx)
                    for el in value
                ],
                ctx=asttrs.Load(),
            )

        if isinstance(value, Set):
            return asttrs.Set(
                elts=[
                    ParamContext.get(el, param_ctx)._target_ast(param_ctx)
                    for el in value
                ],
            )

        if isinstance(value, Dict):  # assume Dict[str, Any]
            return asttrs.Dict(
                keys=[asttrs.Constant(value=k) for k in value.keys()],
                values=[
                    ParamContext.get(el, param_ctx)._target_ast(param_ctx)
                    for el in value.values()
                ],
            )

        if isinstance(value, (MethodType, FunctionType, type)):
            if value.__name__ == "<lambda>":
                raise TypeError("lambda is not supported, please use function instead")

            name = value.__qualname__.split(".")
            name[0] = self.alias

            return asttrs.Name(id=".".join(name), ctx=asttrs.Load())

        if isinstance(value, Literal):
            return value._to_ast()

        raise TypeError(f"{type(value)} is not supported, got: {value}")

    def _dep_ast(self, param_ctx: "ParamContext" = None) -> List[asttrs.stmt]:
        body = []
        value = self.target

        if isinstance(value, (List, Tuple, Set)):
            for el in value:
                body.extend(ParamContext.get(el, param_ctx)._dep_ast(param_ctx))

        elif isinstance(value, Dict):  # assume Dict[str, Any]
            for v in value.values():
                body.extend(ParamContext.get(v, param_ctx)._dep_ast(param_ctx))

        elif isinstance(value, (MethodType, FunctionType, type)):
            if value.__name__ == "<lambda>":
                raise TypeError("lambda is not supported, please use function instead")

            modname = value.__module__
            name = value.__qualname__.split(".")[0]

            body.append(
                asttrs.ImportFrom(
                    module=modname,
                    names=[
                        asttrs.alias(
                            name=name, asname=None if name == self.alias else self.alias
                        )
                    ],
                )
            )

        elif isinstance(value, Literal):

            body.extend(value._dep_ast(param_ctx))

        return body


class ParamContext:

    def __init__(self):
        self.params: Dict[str, Param] = {}
        self.conflicts: Dict[str, int] = {}
        self.aliases: Dict[str, str] = {}

    @classmethod
    def get(cls, obj: Any, param_ctx: "ParamContext" = None) -> Param:
        return param_ctx._get(obj) if param_ctx else Param(obj)

    def _get(self, obj: Any) -> Param:
        if not isinstance(obj, (type, FunctionType, MethodType)):
            if isinstance(obj, (List, Tuple, Set)):
                for el in obj:
                    self._get(el)
            elif isinstance(obj, Dict):
                for el in obj.values():
                    self._get(el)

            return Param(obj)

        elif isinstance(obj, ModuleType):
            # TODO: module doesn't have __qualname__
            raise TypeError(f"ModuleType is not supported, got: {obj}")

        target_name = qualname(obj)
        qual_name = obj.__qualname__
        alias_name = qual_name.split(".")[0]
        import_name = f"{obj.__module__}.{alias_name}"

        if target_name in self.params:
            return self.params[target_name]

        if import_name in self.aliases:
            alias_name = self.aliases[import_name]

        else:
            alias_name = self._find_alias_without_conflict(alias_name)
            self.conflicts[alias_name] = 0
            self.aliases[import_name] = alias_name

        self.params[target_name] = Param(target=obj, alias=alias_name)

        return self.params[target_name]

    def _find_alias_without_conflict(self, name: str) -> str:
        curr, last = name, None
        while curr in self.conflicts:
            last = curr
            curr = f"{curr}_{self.conflicts[curr] + 1}"

        if last:
            self.conflicts[last] += 1

        return curr


class TaskAttribute:
    """
    Class representing a task attribute.

    Attributes:
        op_class (str | Type): The operator class associated with the task attribute. Defaults to "EmptyOperator".
        op_module (str): The module containing the operator class. Defaults to None.
        op_params (Dict[str, Any]): The parameters for the operator. Defaults to None.
        upstream (TaskClass | Iterable[TaskClass]): The upstream task(s) for this attribute. Defaults to None.
        downstream (TaskClass, Iterable[TaskClass]): The downstream task(s) for this attribute. Defaults to None.
    """

    op_class: Union[str, Type] = None
    op_module: Optional[str] = None
    op_params: Dict[str, Any] = None
    upstream: Optional[Union[TaskClass, Iterable[TaskClass]]] = None
    downstream: Optional[Union[TaskClass, Iterable[TaskClass]]] = None


class Task(TaskAttribute):

    @staticmethod
    def _get_taskid(cls) -> str:
        """Use qualified name as task_id
        Provide customized logic by overriding this method, and perhaps `_to_varname` as well.
        Please make sure the returned value is globally unique.
        """

        if (
            hasattr(cls, "_get_taskid")
            and
            # Assume classmethod
            inspect.ismethod(getattr(cls, "_get_taskid"))
        ):
            return cls._get_taskid()
        return qualname(cls)

    @staticmethod
    @lru_cache()
    def _get_attributes(cls) -> TaskAttribute:
        """Get and cache all the task's attributes.

        The expected attribute is defined in `TaskAttribute`,
        each attribute can be assign as a class variable or a @property,
        property precedes class variable if both are defined.

        In order to get the attribute from property, this method creates the task instance based on the Task class, make sure no positional arguments are required to create the task instance.

        """

        self: Union[Task, TaskClass] = None
        try:
            # Assume no arguments are required for creating the task instance.
            self = cls()
        except Exception:
            loguru.logger.warning(
                f"Cannot initialize {cls}, assume attributes are defined in class-var."
            )
            self = cls

        attrs = {}
        for field in TaskAttribute.__annotations__:
            value = getattr(self, field, None)

            if field == "op_class" and value is None:
                raise ValueError("op_class cannot be None")

            attrs[field] = value

        return immutable(TaskAttribute)(**attrs)

    @staticmethod
    def _to_varname(cls) -> str:
        """Represent the Task as a variable name
        The variable name is derived from task_id,
        if you change _get_taskid, you may need to change this method as well.
        """

        if (
            hasattr(cls, "_to_varname")
            and
            # Assume classmethod
            inspect.ismethod(getattr(cls, "_to_varname"))
        ):
            return cls._to_varname()

        return Task._get_taskid(cls).replace(".", "_")

    @staticmethod
    def _collect_dep_ast(cls, param_ctx: ParamContext = None) -> List[asttrs.stmt]:
        """Collect all stmts for all dependencies"""

        vendor_op = Task._resolve_operator(cls)
        op_modname, op_basename = qualname(vendor_op).rsplit(".", 1)
        op_modname = op_modname.replace("airfly._vendor.", "")

        deps = [
            asttrs.ImportFrom(module=op_modname, names=[asttrs.alias(name=op_basename)])
        ]

        params = Task._get_attributes(cls).op_params

        deps.extend(ParamContext.get(params, param_ctx)._dep_ast(param_ctx))

        return deps

    @staticmethod
    @lru_cache()
    def _collect_op_annotations(op: Type) -> Dict[str, Any]:
        """Collect annotations for the operator."""
        avai_params = {}

        if Task._is_builtin_op(op):
            for base in op.mro()[::-1]:
                avai_params.update(getattr(base, "__annotations__", {}))

        elif Task._is_private_op(op):
            for base in op.mro()[::-1]:
                signature = inspect.signature(base)
                for k, v in signature.parameters.items():

                    anno = v.annotation
                    if isinstance(anno, type):
                        typename = qualname(anno, level=1)
                    elif hasattr(anno, "__module__") and anno.__module__ == "typing":
                        typename = str(anno)
                    elif isinstance(anno, str):
                        typename = anno
                    else:
                        raise TypeError(f"got: {anno}")

                    if v.kind not in [
                        inspect._ParameterKind.VAR_KEYWORD,
                        inspect._ParameterKind.VAR_POSITIONAL,
                    ]:

                        avai_params.update({k: typename})
        else:
            raise ValueError(f"Operator {op} is not supported.")

        return avai_params

    @staticmethod
    def _to_ast(
        cls, param_ctx: ParamContext = None, task_group: bool = True
    ) -> asttrs.AST:
        """
        Generate an Abstract Syntax Tree (AST) representation of the Task.

        Returns:
            asttrs.AST: The AST representation of the Task.

        Notes:
            - This method uses the `asttrs` module for creating the AST nodes.
            - The AST represents the assignment of a Task instance to a variable.
            - The Task instance is created by calling the resolved operator class with the appropriate parameters.
            - The parameters are obtained from the Task's attributes and the operator's available parameters.
            - The AST node is an assignment statement, where the target is the variable name and the value is the function call.

        # TODO: Add unit tests for this method.
        """

        op = Task._resolve_operator(cls)
        op_basename = qualname(op, level=1)

        task_id = Task._get_taskid(cls)
        task_varname = Task._to_varname(cls)

        avai_params = Task._collect_op_annotations(op)
        params = dict(task_id=task_id)

        for k, v in (Task._get_attributes(cls).op_params or {}).items():
            if k in params:
                loguru.logger.warning(f"Overwrite param {k}={params[k]} with value {v}")
            if k in avai_params:
                params.update({k: v})
            else:
                loguru.logger.warning(f"Ingore invalid param: {k} in {cls}.")

        keywords = []

        for k, v in params.items():
            par = ParamContext.get(v, param_ctx)
            keywords.append(asttrs.keyword(arg=k, value=par._target_ast(param_ctx)))

        if task_group:
            group_id = task_id.rsplit(".", 1)[0]
            group_var = TaskGroup._to_varname(group_id)
            keywords.append(
                asttrs.keyword(
                    arg="task_group",
                    value=asttrs.Name(id=group_var, ctx=asttrs.Load()),
                )
            )

        assign = asttrs.Assign(
            targets=[asttrs.Name(id=task_varname, ctx=asttrs.Store())],
            value=asttrs.Call(
                func=asttrs.Name(id=op_basename, ctx=asttrs.Load()),
                keywords=keywords,
            ),
        )

        return assign

    @staticmethod
    @lru_cache()
    def _is_builtin_op(op: Type) -> bool:
        return issubclass_by_qualname(op, BUILTIN_OPERATORS["BaseOperator"])

    @staticmethod
    @lru_cache()
    def _is_private_op(op: Type) -> bool:
        try:
            from airflow.models.baseoperator import BaseOperator

            return issubclass_by_qualname(op, BaseOperator)
        except Exception as err:
            loguru.logger.warning(f"{err.__class__}: {err}")
            pass

        return False

    @staticmethod
    def _resolve_operator(cls) -> Type:
        """Resolve the operator class for the Task.

        Returns:
            Type: The resolved operator class.

        Raises:
            ValueError: If the op_class is invalid or not found in the BUILTIN_OPERATORS.
            ValueError: If multiple op_class with the same basename are found and op_module is not provided or invalid.
            ValueError: If the op_class candidates cannot be resolved by the given op_module.
        """

        op_class = Task._get_attributes(cls).op_class

        if isinstance(op_class, type):

            if Task._is_builtin_op(op_class) or Task._is_private_op(op_class):
                return op_class

            raise TypeError(f"Not a valid operator type, got: {op_class}")

        # NOTE: if op_class is str, assume it's a builtin operator
        if isinstance(op_class, str):
            basename = op_class

        else:
            raise ValueError(f"Invalid op_class, got: {op_class}")

        if basename not in BUILTIN_OPERATORS:
            raise ValueError(
                f"'{basename}' not found. If this is unexpected and the operator should exist, please report the issue."
            )
        items = BUILTIN_OPERATORS[basename]

        if len(items) > 1:
            # Disambiguate by op_module
            op_module = Task._get_attributes(cls).op_module
            if not (op_module and isinstance(op_module, str)):
                raise ValueError(
                    f"Multiple op_class('{basename}') found: {items}, but op_module is invalid for resolving the operator, got: {op_module}"
                )

            cands = [op for op in items if op_module in qualname(op)]

            if len(cands) != 1:
                raise ValueError(
                    f"Cannot resolve the possible op_class('{basename}') by given op_module('{op_module}'), multiple or no candidates found: {cands}"
                )

            items = cands

        return items[0]


class AirFly(Task): ...


@immutable
class TaskPair:
    """
    A class representing a pair of tasks.

    Attributes:
        up (TaskClass): The class representing the 'up' task.
        down (TaskClass): The class representing the 'down' task.

    """

    up: TaskClass
    down: TaskClass

    def _to_ast(self) -> asttrs.AST:
        """Returns an Abstract Syntax Tree (AST) representation of the TaskPair.

        i.e., an AST to represent an expression: "up_task >> down_task"

        This method generates an AST that represents the TaskPair as a binary operation expression. The left operand of the binary operation is the result of calling the '_to_varname' method on the 'up' task, and the right operand is the result of calling the '_to_varname' method on the 'down' task.

        Returns:
            asttrs.AST: The AST representation of the TaskPair.

        """

        return asttrs.Expr(
            value=asttrs.BinOp(
                left=asttrs.Name(id=Task._to_varname(self.up), ctx=asttrs.Store()),
                op=asttrs.RShift(),
                right=asttrs.Name(id=Task._to_varname(self.down), ctx=asttrs.Load()),
            )
        )


@immutable
class TaskGroup:
    group_id: str
    parent_id: str = None

    def _to_ast(self) -> asttrs.stmt:
        keywords = [
            asttrs.keyword(arg="group_id", value=asttrs.Constant(value=self.group_id)),
            asttrs.keyword(arg="prefix_group_id", value=asttrs.Constant(value=False)),
        ]

        if self.parent_id:
            keywords.append(
                asttrs.keyword(
                    arg="parent_group",
                    value=asttrs.Name(
                        id=self._to_varname(self.parent_id), ctx=asttrs.Load()
                    ),
                )
            )

        return asttrs.Assign(
            targets=[
                asttrs.Name(id=self._to_varname(self.group_id), ctx=asttrs.Store())
            ],
            value=asttrs.Call(
                func=asttrs.Name(id="TaskGroup", ctx=asttrs.Load()),
                keywords=keywords,
            ),
        )

    @classmethod
    def _to_varname(cls, group_id) -> str:
        return "group_" + group_id.replace(".", "_")


@immutable
class TaskTree:
    """A class representing a task tree.

    The TaskTree class represents a task tree, which consists of a set of task classes and a set of task pairs. The task classes are collected from a module based on certain criteria, such as being a subclass of a base task class and satisfying a predicate function. The task pairs represent the dependencies between tasks.

    Attributes:
        taskset (Set[TaskClass]): The set of task classes in the task tree.
        taskpairs (Set[TaskPair]): The set of task pairs in the task tree.

    Methods:
        from_module(module: ModuleType) -> TaskTree:
            Creates a task tree from a module.
            This method takes a module as input and creates a task tree from it. The task tree consists of a set of task classes and a set of task pairs. The task classes are collected from the module based on certain criteria, such as being a subclass of a base task class and satisfying a predicate function. The task pairs are generated from the task classes, representing the dependency between tasks.

        to_source() -> str:
            Generates the source code representation of the task tree.
            This method generates the source code representation of the task tree, including the import statements, class definition, and task pair expressions.

    Example:
    >>> TaskTree.from_module(module).to_source(formatted=True)

    Note:
        - The TaskTree class is immutable, meaning that its attributes cannot be modified after initialization.
        - The taskset attribute is a set of task classes, where each task class represents a task in the task tree.
        - The taskpairs attribute is a set of task pairs, where each task pair represents a dependency between two tasks in the task tree.
        - The _dag attribute is a directed graph (DiGraph) object from the NetworkX library, which represents the task tree as a directed acyclic graph (DAG).
        - The from_module method is a class method that creates a task tree from a module.
        - The to_source method generates the source code representation of the task tree, which can be used to create a Python script or module.
    """

    taskset: Set[TaskClass]
    taskpairs: Set[TaskPair]

    _dag: nx.DiGraph = attr.ib(init=False)

    @_dag.default
    def _create_dag(self):
        dag = nx.DiGraph()
        dag.add_nodes_from(self.taskset)
        dag.add_edges_from((pair.up, pair.down) for pair in self.taskpairs)

        return dag

    @classmethod
    def from_module(
        cls, module: ModuleType, taskclass: type = AirFly, exclude_pattern: str = None
    ) -> "TaskTree":
        """
        Creates a task tree from a module.

        This method takes a module as input and creates a task tree from it. The task tree consists of a set of task classes and a set of task pairs. The task classes are collected from the module based on certain criteria, such as being a subclass of a base task class and satisfying a predicate function. The task pairs are generated from the task classes, representing the dependency between tasks.

        Parameters:
            module (ModuleType): The module from which to create the task tree.

        Returns:
            TaskTree: The task tree created from the module.

        Example:
            >>> module = ...
            >>> task_tree = TaskTree.from_module(module)

        # TODO: add predicate in argument
        """

        loguru.logger.info(
            f"Collecting '{qualname(taskclass)}' from '{qualname(module)}' ..."
        )
        taskset = set(
            cls._collect_taskclass(
                module,
                taskclass,
                predicate=lambda obj: not cls._should_exclude(obj, exclude_pattern),
            )
        )
        taskpairs = set(
            cls._collect_taskpairs(
                taskset,
                taskclass,
                predicate=lambda pair: not (
                    cls._should_exclude(
                        pair.up,
                        exclude_pattern,
                    )
                    or cls._should_exclude(
                        pair.down,
                        exclude_pattern,
                    )
                ),
            )
        )

        return cls(taskset=taskset, taskpairs=taskpairs)

    @classmethod
    def _should_exclude(
        cls,
        obj: Union[FunctionType, type],
        pattern: Optional[str] = None,
    ):
        if pattern and isinstance(pattern, str):
            obj_name = qualname(obj)
            if re.search(pattern, obj_name):
                return True

        return False

    @classmethod
    def _collect_taskclass(
        cls, module: ModuleType, taskclass: TaskClass = Task, predicate=lambda _: True
    ) -> Generator[TaskClass, None, None]:
        """
        Collects task classes from a module.

        This method collects task classes from a given module based on the provided criteria. It returns a generator that yields the task classes.

        Parameters:
            module (ModuleType): The module from which to collect the task classes.
            taskclass (Type["Task"], optional): The base task class to filter the task classes. Defaults to Task.
            predicate (Callable[[Union[FunctionType, type]], bool], optional): A predicate function to further filter the task classes. Defaults to lambda _: True.

        Yields:
            TaskClass: The task classes that satisfy the filtering criteria.

        Example:
            >>> module = ...
            >>> task_classes = list(TaskTree._collect_taskclass(module))
        """
        for cls in collect_objects(
            module,
            predicate=lambda obj: isinstance(obj, type)
            and issubclass_by_qualname(obj, taskclass)
            and predicate(obj),
        ):

            yield cls

    @classmethod
    def _collect_taskpairs(
        cls,
        taskset: Set[TaskClass],
        taskclass: TaskClass = Task,
        predicate=lambda _: True,
    ) -> Generator[TaskPair, None, None]:
        """
        Collects task pairs from a given task set.

        This method takes a set of task classes and collects all possible pairs of tasks based on the provided criteria. It returns a generator that yields the task pairs.

        Parameters:
            taskset (Set[TaskClass]): The set of task classes from which to collect the task pairs.
            taskclass (TaskClass): The base task class to filter the task pairs. Defaults to Task.
            predicate (Callable[[TaskPair], bool], optional): A predicate function to further filter the task pairs. Defaults to lambda _: True.

        Yields:
            TaskPair: The task pairs that satisfy the filtering criteria.

        Example:
            >>> taskset = {Task1, Task2, Task3}
            >>> task_pairs = list(TaskTree._collect_taskpairs(taskset))
        """

        cached = set()

        for cls in taskset:
            ups = Task._get_attributes(
                cls
            ).upstream  # could be None, Task or Iterable[Task, ...]

            if isinstance(ups, type) and issubclass_by_qualname(ups, taskclass):
                ups = [ups]

            for u in ups or []:

                if isinstance(u, type) and issubclass_by_qualname(u, taskclass):
                    pair = TaskPair(up=u, down=cls)

                    if (pair not in cached) and predicate(pair):
                        cached.add(pair)

                        yield pair

            downs = Task._get_attributes(
                cls
            ).downstream  # could be None, Task or Tuple[Task, ...]

            if isinstance(downs, type) and issubclass_by_qualname(downs, taskclass):
                downs = [downs]

            for d in downs or []:

                if isinstance(d, type) and issubclass_by_qualname(d, taskclass):
                    pair = TaskPair(up=cls, down=d)

                    if pair not in cached and predicate(pair):
                        cached.add(pair)

                        yield pair

    def to_dag(
        self,
        name: str,
        includes: Union[str, List[str]] = None,
        dag_params: Tuple[str, str] = None,
        task_group: bool = True,
        formatted: bool = True,
    ) -> str:
        """Generates the source code representation of the task tree.

        This method generates the source code representation of the task tree by converting the abstract syntax tree (AST) of the task tree to source code. The AST is obtained by calling the `_to_ast` method. The source code is then passed through the `blacking` and `isorting` functions to format and sort the code, respectively.

        Parameters:
            formatted (bool, optional): Specifies whether the generated source code should be formatted. If `True`, the code will be formatted using the `blacking` and `isorting` functions. If `False`, the code will be returned as is. Defaults to `True`.

        Returns:
            str: The source code representation of the task tree.

        Note:
            - The `formatted` parameter determines whether the generated source code should be formatted. If `True`, the code will be formatted using the `blacking` and `isorting` functions. If `False`, the code will be returned as is.
            - The `blacking` function is used to format the code by applying the `black` code formatter.
            - The `isorting` function is used to sort the imports in the code by applying the `isort` import sorter.
            - The `re.sub` function is used to remove any consecutive newline characters in the source code.

        Example:
            >>> task_tree = TaskTree(...)
            >>> source_code = task_tree.to_source(formatted=True)
        """

        loguru.logger.info(
            f"Generating code of dag.py: name='{name}', includes={includes}, dag_params={dag_params}"
        )

        src = re.sub(
            "\n+",
            "\n",
            self._to_ast(name, dag_params, includes, task_group).to_source(),
        )

        return isorting(blacking(src)) if formatted else src

    def _to_ast(
        self,
        dag_name,
        dag_params,
        includes,
        task_group=True,
    ) -> asttrs.AST:

        ctx = ParamContext()
        body = (
            self._build_header()
            + self._build_imports(task_group)
            + self._build_includes(includes)
            + self._build_dag_context(
                dag_name=dag_name,
                dag_params=dag_params,
                param_ctx=ctx,
                task_group=task_group,
            )
        )

        return asttrs.Module(body=body)

    def _build_header(self) -> List[asttrs.stmt]:

        return [
            asttrs.Comment(
                body=f"This file is auto-generated by {airfly.__name__} {airfly.__version__}"
            )
        ]

    def _build_imports(self, task_group=True) -> List[asttrs.stmt]:

        imports = [
            asttrs.ImportFrom(module="airflow.models", names=[asttrs.alias(name="DAG")])
        ]

        if task_group:
            imports.append(
                asttrs.ImportFrom(
                    module="airflow.utils.task_group",
                    names=[asttrs.alias(name="TaskGroup")],
                )
            )

        return imports

    def _build_includes(
        self, includes: Union[str, List[str]] = None
    ) -> List[asttrs.stmt]:
        if includes:
            if isinstance(includes, str):
                includes = [includes]

            imports = []
            statements = []

            for path in includes:
                for st in self._insert_from_pyfile(path):
                    if isinstance(st, (asttrs.Import, asttrs.ImportFrom)):
                        if st not in imports:
                            imports.append(st)
                    else:
                        statements.append(st)

            body = (
                imports
                + statements
                + [asttrs.Comment(body="<" * 10 + " End of code insertion")]
            )

            return body

        return []

    def _insert_from_pyfile(self, path: str) -> List[asttrs.stmt]:

        try:

            if os.path.isfile(path):
                mod: asttrs.Module = asttrs.Module.from_file(path)
                mod.body.insert(
                    0, asttrs.Comment(body=">" * 10 + f" Include from '{path}'")
                )

                return mod.body

        except Exception as e:
            loguru.logger.warning(f"Ignore failed insert from '{path}': {e}")
            pass

        return []

    def _build_dag_context(
        self,
        dag_name: str,
        dag_params: Tuple[str, str] = None,
        param_ctx=None,
        task_group=True,
    ) -> List[asttrs.stmt]:

        if dag_params and dag_params[1]:
            _, param_var = dag_params
            keywords = [
                asttrs.keyword(
                    arg=None,
                    value=asttrs.Name(id=param_var, ctx=asttrs.Load()),
                )
            ]
        else:
            keywords = []

        return [
            asttrs.With(
                items=[
                    asttrs.withitem(
                        context_expr=asttrs.Call(
                            func=asttrs.Name(id="DAG", ctx=asttrs.Load()),
                            args=[asttrs.Constant(value=dag_name)],
                            keywords=keywords,
                        ),
                        optional_vars=asttrs.Name(id="dag", ctx=asttrs.Store()),
                    )
                ],
                body=self._build_dag_body(param_ctx=param_ctx, task_group=task_group),
            )
        ]

    def _build_dag_body(self, param_ctx=None, task_group=True) -> List[asttrs.stmt]:
        body = []

        for cls in self.taskset:
            for st in Task._collect_dep_ast(cls, param_ctx):
                if st not in body:
                    body.append(st)

        if task_group:
            body.extend(self._build_task_group())

        for cls in sorted(self.taskset, key=lambda el: qualname(el)):
            body.append(Task._to_ast(cls, param_ctx, task_group))

        for pair in sorted(
            self.taskpairs,
            key=lambda el: f"{qualname(el.up)}{qualname(el.down)}",
        ):

            body.append(pair._to_ast())

        return body if body else [asttrs.Pass()]

    def _build_task_group(self) -> List[asttrs.stmt]:

        body = []
        group_ids = sorted(
            {Task._get_taskid(cls).rsplit(".", 1)[0] for cls in self.taskset}
        )

        tree = {}
        for gid in group_ids:
            node = tree
            for path in gid.split("."):
                if path not in node:
                    node[path] = {}
                node = node[path]

        todo = deque([(tree, [])])
        while todo:
            root, prefix = todo.popleft()

            for k, node in root.items():
                gid = ".".join(prefix + [k])
                body.append(
                    TaskGroup(
                        group_id=gid, parent_id=".".join(prefix) if prefix else None
                    )._to_ast()
                )

                todo.append((node, prefix + [k]))

        return body



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.sensors.base import BaseSensorOperator


class DataplexTaskStateSensor(BaseSensorOperator):
    project_id: "str"
    region: "str"
    lake_id: "str"
    dataplex_task_id: "str"
    api_version: "str"
    retry: "Retry | _MethodDefault"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class DataplexDataQualityJobStatusSensor(BaseSensorOperator):
    project_id: "str"
    region: "str"
    data_scan_id: "str"
    job_id: "str"
    api_version: "str"
    retry: "Retry | _MethodDefault"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"
    fail_on_dq_failure: "bool"
    result_timeout: "float"
    start_sensor_time: "float | None"


class DataplexDataProfileJobStatusSensor(BaseSensorOperator):
    project_id: "str"
    region: "str"
    data_scan_id: "str"
    job_id: "str"
    api_version: "str"
    retry: "Retry | _MethodDefault"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"
    result_timeout: "float"
    start_sensor_time: "float | None"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.models.baseoperator import BaseOperator


class DatabricksReposCreateOperator(BaseOperator):
    git_url: "str"
    git_provider: "str | None"
    branch: "str | None"
    tag: "str | None"
    repo_path: "str | None"
    ignore_existing_repo: "bool"
    databricks_conn_id: "str"
    databricks_retry_limit: "int"
    databricks_retry_delay: "int"


class DatabricksReposUpdateOperator(BaseOperator):
    branch: "str | None"
    tag: "str | None"
    repo_id: "str | None"
    repo_path: "str | None"
    databricks_conn_id: "str"
    databricks_retry_limit: "int"
    databricks_retry_delay: "int"


class DatabricksReposDeleteOperator(BaseOperator):
    repo_id: "str | None"
    repo_path: "str | None"
    databricks_conn_id: "str"
    databricks_retry_limit: "int"
    databricks_retry_delay: "int"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.models.baseoperator import BaseOperator


class GoogleDisplayVideo360CreateQueryOperator(BaseOperator):
    body: "dict[str, Any]"
    api_version: "str"
    gcp_conn_id: "str"
    delegate_to: "str | None"
    impersonation_chain: "str | Sequence[str] | None"


class GoogleDisplayVideo360CreateSDFDownloadTaskOperator(BaseOperator):
    body_request: "dict[str, Any]"
    api_version: "str"
    gcp_conn_id: "str"
    delegate_to: "str | None"
    impersonation_chain: "str | Sequence[str] | None"


class GoogleDisplayVideo360DeleteReportOperator(BaseOperator):
    report_id: "str | None"
    report_name: "str | None"
    api_version: "str"
    gcp_conn_id: "str"
    delegate_to: "str | None"
    impersonation_chain: "str | Sequence[str] | None"


class GoogleDisplayVideo360DownloadLineItemsOperator(BaseOperator):
    request_body: "dict[str, Any]"
    bucket_name: "str"
    object_name: "str"
    gzip: "bool"
    api_version: "str"
    gcp_conn_id: "str"
    delegate_to: "str | None"
    impersonation_chain: "str | Sequence[str] | None"


class GoogleDisplayVideo360DownloadReportV2Operator(BaseOperator):
    query_id: "str"
    report_id: "str"
    bucket_name: "str"
    report_name: "str | None"
    gzip: "bool"
    chunk_size: "int"
    api_version: "str"
    gcp_conn_id: "str"
    delegate_to: "str | None"
    impersonation_chain: "str | Sequence[str] | None"


class GoogleDisplayVideo360RunQueryOperator(BaseOperator):
    query_id: "str"
    parameters: "dict[str, Any] | None"
    api_version: "str"
    gcp_conn_id: "str"
    delegate_to: "str | None"
    impersonation_chain: "str | Sequence[str] | None"


class GoogleDisplayVideo360SDFtoGCSOperator(BaseOperator):
    operation_name: "str"
    bucket_name: "str"
    object_name: "str"
    gzip: "bool"
    api_version: "str"
    gcp_conn_id: "str"
    delegate_to: "str | None"
    impersonation_chain: "str | Sequence[str] | None"


class GoogleDisplayVideo360UploadLineItemsOperator(BaseOperator):
    bucket_name: "str"
    object_name: "str"
    api_version: "str"
    gcp_conn_id: "str"
    delegate_to: "str | None"
    impersonation_chain: "str | Sequence[str] | None"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.models.baseoperator import BaseOperator


class BashOperator(BaseOperator):
    bash_command: "str | ArgNotSet"
    env: "dict[str, str] | None"
    append_env: "bool"
    output_encoding: "str"
    skip_exit_code: "int | None"
    skip_on_exit_code: "int | Container[int] | None"
    cwd: "str | None"
    output_processor: "Callable[[str], Any]"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.providers.amazon.aws.operators.base_aws import (
    AwsBaseOperator,
)


class SqsPublishOperator(AwsBaseOperator):
    sqs_queue: "str"
    message_content: "str"
    message_attributes: "dict | None"
    delay_seconds: "int"
    message_group_id: "str | None"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.models.baseoperator import BaseOperator


class LocalFilesystemToWasbOperator(BaseOperator):
    file_path: "str"
    container_name: "str"
    blob_name: "str"
    wasb_conn_id: "str"
    create_container: "bool"
    load_options: "dict | None"



import importlib
import io
import os
import pathlib
import pkgutil
import subprocess as sp
import sys
import tempfile
from contextlib import contextmanager
from types import FunctionType, ModuleType
from typing import Callable, Generator, Union

import attr
import loguru

immutable = attr.s(auto_attribs=True, slots=True, frozen=True, kw_only=True)


def qualname(obj: Union[FunctionType, ModuleType, type], level: int = -1) -> str:
    """Return the qualname of a class, a function or a module.

    >>> class Foo:
    ...     pass

    >>> def bar():
    ...     pass

    >>> qualname(Foo)
    'airfly.utils.Foo'

    >>> qualname(bar)
    'airfly.utils.bar'
    """

    if isinstance(obj, FunctionType) or isinstance(obj, type):
        name = f"{obj.__module__}.{obj.__qualname__}"

    elif isinstance(obj, ModuleType):
        name = f"{obj.__name__}"

    else:
        name = ""

    return ".".join(name.split(".")[-level:]) if level > 0 else name


def issubclass_by_qualname(cls, class_or_tuple):
    mro = {qualname(o) for o in cls.mro()}

    if isinstance(class_or_tuple, type):
        class_or_tuple = (class_or_tuple,)

    for o in class_or_tuple:
        if qualname(o) in mro:
            return True

    return False


def blacking(source_code: str):

    with tempfile.NamedTemporaryFile("w", delete=False) as f:
        f.write(source_code)
        fname = f.name

    with sp.Popen(["cat", fname], stdout=sp.PIPE) as p:
        cmd = [sys.executable, "-m", "black", "-q", "-"]
        out = sp.check_output(cmd, stdin=p.stdout)

    try:
        pathlib.Path(fname).unlink()

    except FileNotFoundError:
        pass

    return out.decode()


def isorting(source_code: str):

    with tempfile.NamedTemporaryFile("w", delete=False) as f:
        f.write(source_code)
        fname = f.name

    with sp.Popen(["cat", fname], stdout=sp.PIPE) as p:
        cmd = [sys.executable, "-m", "isort", "--profile", "black", "-q", "-"]
        out = sp.check_output(cmd, stdin=p.stdout)

    try:
        pathlib.Path(fname).unlink()

    except FileNotFoundError:
        pass

    return out.decode()


def collect_objects(
    module: ModuleType,
    predicate: Callable[[Union[FunctionType, type]], bool] = lambda _: True,
) -> Generator[Union[FunctionType, type], None, None]:

    modname = qualname(module)
    cached = set()

    impls = (
        _collect_from_package(module)
        if hasattr(
            module, "__path__"
        )  # By definition, if a module has a __path__ attribute, it is a package.
        else _collect_from_module(module)
    )

    for obj in impls:
        name = qualname(obj)
        if obj.__module__.startswith(modname) and (name not in cached):
            if predicate(obj):
                cached.add(name)
                yield obj


@contextmanager
def _escape_any_commandline_parser():

    argv = sys.argv
    stdout = sys.stdout
    stderr = sys.stderr

    sys.argv = [""]
    sys.stdout = io.StringIO()
    sys.stderr = io.StringIO()

    yield

    sys.argv = argv
    sys.stdout = stdout
    sys.stderr = stderr


def _collect_from_package(
    package: ModuleType,
) -> Generator[Union[FunctionType, type], None, None]:

    for obj in _collect_from_module(package):
        yield obj

    for _, name, _ in pkgutil.walk_packages(package.__path__, package.__name__ + "."):

        with _escape_any_commandline_parser():
            try:
                mod = importlib.import_module(name)

            except Exception as e:
                loguru.logger.warning(f"Ignore invalid module: '{name}'. Reason: {e}")
                continue

        for obj in _collect_from_module(mod):
            yield obj


def _collect_from_module(
    module: ModuleType,
) -> Generator[Union[FunctionType, type], None, None]:

    for key, obj in module.__dict__.items():

        try:
            if isinstance(obj, (type, FunctionType)):
                yield obj
        except Exception as err:
            loguru.logger.warning(f"Ignore invalid object: '{key}'. Reason: {err}")


def makefile(fullpath: str, content: Union[str, bytes] = "", overwrite: bool = False):

    if not os.path.isfile(fullpath):
        dirpath = os.path.dirname(fullpath)

        if not os.path.isdir(dirpath):
            os.makedirs(dirpath)

        _writefile(fullpath, content)

    elif overwrite:
        _writefile(fullpath, content)


def _writefile(fullpath: str, content: Union[str, bytes] = ""):

    if not isinstance(content, (str, bytes)):
        raise TypeError(f"content must be str or bytes, got: {type(content)}")

    with open(fullpath, "wb" if isinstance(content, bytes) else "w") as f:
        f.write(content)



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.models.baseoperator import BaseOperator


class LivyOperator(BaseOperator):
    file: "str"
    class_name: "str | None"
    args: "Sequence[str | int | float] | None"
    conf: "dict[Any, Any] | None"
    jars: "Sequence[str] | None"
    py_files: "Sequence[str] | None"
    files: "Sequence[str] | None"
    driver_memory: "str | None"
    driver_cores: "int | str | None"
    executor_memory: "str | None"
    executor_cores: "int | str | None"
    num_executors: "int | str | None"
    archives: "Sequence[str] | None"
    queue: "str | None"
    name: "str | None"
    proxy_user: "str | None"
    livy_conn_id: "str"
    livy_conn_auth_type: "Any | None"
    polling_interval: "int"
    extra_options: "dict[str, Any] | None"
    extra_headers: "dict[str, Any] | None"
    retry_args: "dict[str, Any] | None"
    deferrable: "bool"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.models.baseoperator import BaseOperator


class AQLOperator(BaseOperator):
    query: "str"
    arangodb_conn_id: "str"
    result_processor: "Callable | None"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.providers.google.cloud.operators.cloud_base import (
    GoogleCloudBaseOperator,
)


class DataprocCreateClusterOperator(GoogleCloudBaseOperator):
    cluster_name: "str"
    region: "str"
    project_id: "str"
    cluster_config: "dict | Cluster | None"
    virtual_cluster_config: "dict | None"
    labels: "dict | None"
    request_id: "str | None"
    delete_on_error: "bool"
    use_if_exists: "bool"
    retry: "AsyncRetry | _MethodDefault"
    timeout: "float"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"
    deferrable: "bool"
    polling_interval_seconds: "int"


class DataprocScaleClusterOperator(GoogleCloudBaseOperator):
    pass


class DataprocDeleteClusterOperator(GoogleCloudBaseOperator):
    region: "str"
    cluster_name: "str"
    project_id: "str"
    cluster_uuid: "str | None"
    request_id: "str | None"
    retry: "AsyncRetry | _MethodDefault"
    timeout: "float"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"
    deferrable: "bool"
    polling_interval_seconds: "int"


class _DataprocStartStopClusterBaseOperator(GoogleCloudBaseOperator):
    cluster_name: "str"
    region: "str"
    project_id: "str"
    cluster_uuid: "str | None"
    request_id: "str | None"
    retry: "AsyncRetry | _MethodDefault"
    timeout: "float"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class DataprocStartClusterOperator(_DataprocStartStopClusterBaseOperator):
    cluster_name: "str"
    region: "str"
    project_id: "str"
    cluster_uuid: "str | None"
    request_id: "str | None"
    retry: "AsyncRetry | _MethodDefault"
    timeout: "float"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class DataprocStopClusterOperator(_DataprocStartStopClusterBaseOperator):
    cluster_name: "str"
    region: "str"
    project_id: "str"
    cluster_uuid: "str | None"
    request_id: "str | None"
    retry: "AsyncRetry | _MethodDefault"
    timeout: "float"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class DataprocJobBaseOperator(GoogleCloudBaseOperator):
    region: "str"
    job_name: "str"
    cluster_name: "str"
    project_id: "str"
    dataproc_properties: "dict | None"
    dataproc_jars: "list[str] | None"
    gcp_conn_id: "str"
    labels: "dict | None"
    job_error_states: "set[str] | None"
    impersonation_chain: "str | Sequence[str] | None"
    asynchronous: "bool"
    deferrable: "bool"
    polling_interval_seconds: "int"


class DataprocSubmitPigJobOperator(DataprocJobBaseOperator):
    pass


class DataprocSubmitHiveJobOperator(DataprocJobBaseOperator):
    pass


class DataprocSubmitSparkSqlJobOperator(DataprocJobBaseOperator):
    pass


class DataprocSubmitSparkJobOperator(DataprocJobBaseOperator):
    pass


class DataprocSubmitHadoopJobOperator(DataprocJobBaseOperator):
    pass


class DataprocSubmitPySparkJobOperator(DataprocJobBaseOperator):
    pass


class DataprocCreateWorkflowTemplateOperator(GoogleCloudBaseOperator):
    template: "dict"
    region: "str"
    project_id: "str"
    retry: "Retry | _MethodDefault"
    timeout: "float | None"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class DataprocInstantiateWorkflowTemplateOperator(GoogleCloudBaseOperator):
    template_id: "str"
    region: "str"
    project_id: "str"
    version: "int | None"
    request_id: "str | None"
    parameters: "dict[str, str] | None"
    retry: "AsyncRetry | _MethodDefault"
    timeout: "float | None"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"
    deferrable: "bool"
    polling_interval_seconds: "int"
    cancel_on_kill: "bool"


class DataprocInstantiateInlineWorkflowTemplateOperator(GoogleCloudBaseOperator):
    template: "dict"
    region: "str"
    project_id: "str"
    request_id: "str | None"
    retry: "AsyncRetry | _MethodDefault"
    timeout: "float | None"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"
    deferrable: "bool"
    polling_interval_seconds: "int"
    cancel_on_kill: "bool"


class DataprocSubmitJobOperator(GoogleCloudBaseOperator):
    job: "dict"
    region: "str"
    project_id: "str"
    request_id: "str | None"
    retry: "Retry | _MethodDefault"
    timeout: "float | None"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"
    asynchronous: "bool"
    deferrable: "bool"
    polling_interval_seconds: "int"
    cancel_on_kill: "bool"
    wait_timeout: "int | None"


class DataprocUpdateClusterOperator(GoogleCloudBaseOperator):
    cluster_name: "str"
    cluster: "dict | Cluster"
    update_mask: "dict | FieldMask"
    graceful_decommission_timeout: "dict | Duration"
    region: "str"
    request_id: "str | None"
    project_id: "str"
    retry: "AsyncRetry | _MethodDefault"
    timeout: "float | None"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"
    deferrable: "bool"
    polling_interval_seconds: "int"


class DataprocDiagnoseClusterOperator(GoogleCloudBaseOperator):
    region: "str"
    cluster_name: "str"
    project_id: "str"
    tarball_gcs_dir: "str | None"
    diagnosis_interval: "dict | Interval | None"
    jobs: "MutableSequence[str] | None"
    yarn_application_ids: "MutableSequence[str] | None"
    retry: "AsyncRetry | _MethodDefault"
    timeout: "float"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"
    deferrable: "bool"
    polling_interval_seconds: "int"


class DataprocCreateBatchOperator(GoogleCloudBaseOperator):
    region: "str | None"
    project_id: "str"
    batch: "dict | Batch"
    batch_id: "str"
    request_id: "str | None"
    retry: "Retry | _MethodDefault"
    timeout: "float | None"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"
    result_retry: "AsyncRetry | _MethodDefault | Retry"
    asynchronous: "bool"
    deferrable: "bool"
    polling_interval_seconds: "int"


class DataprocDeleteBatchOperator(GoogleCloudBaseOperator):
    batch_id: "str"
    region: "str"
    project_id: "str"
    retry: "Retry | _MethodDefault"
    timeout: "float | None"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class DataprocGetBatchOperator(GoogleCloudBaseOperator):
    batch_id: "str"
    region: "str"
    project_id: "str"
    retry: "Retry | _MethodDefault"
    timeout: "float | None"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"


class DataprocListBatchesOperator(GoogleCloudBaseOperator):
    region: "str"
    project_id: "str"
    page_size: "int | None"
    page_token: "str | None"
    retry: "Retry | _MethodDefault"
    timeout: "float | None"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"
    filter: "str | None"
    order_by: "str | None"


class DataprocCancelOperationOperator(GoogleCloudBaseOperator):
    operation_name: "str"
    region: "str"
    project_id: "str"
    retry: "Retry | _MethodDefault"
    timeout: "float | None"
    metadata: "Sequence[tuple[str, str]]"
    gcp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.sensors.base import BaseSensorOperator


class HdfsSensor(BaseSensorOperator):
    pass


class HdfsRegexSensor(HdfsSensor):
    pass


class HdfsFolderSensor(HdfsSensor):
    pass



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.models.baseoperator import BaseOperator


class CopyFromExternalStageToSnowflakeOperator(BaseOperator):
    files: "list | None"
    table: "str"
    stage: "str"
    prefix: "str | None"
    file_format: "str"
    schema: "str | None"
    columns_array: "list | None"
    pattern: "str | None"
    warehouse: "str | None"
    database: "str | None"
    autocommit: "bool"
    snowflake_conn_id: "str"
    role: "str | None"
    authenticator: "str | None"
    session_parameters: "dict | None"
    copy_options: "str | None"
    validation_mode: "str | None"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.models.baseoperator import BaseOperator


class LocalFilesystemToADLSOperator(BaseOperator):
    local_path: "str"
    remote_path: "str"
    overwrite: "bool"
    nthreads: "int"
    buffersize: "int"
    blocksize: "int"
    extra_upload_options: "dict[str, Any] | None"
    azure_data_lake_conn_id: "str"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.models.baseoperator import BaseOperator
from airfly._vendor.airflow.models.skipmixin import SkipMixin


class BranchMixIn(SkipMixin):
    context: "_empty"


class BaseBranchOperator(BaseOperator, BranchMixIn):
    task_id: "str"
    owner: "str"
    email: "str | Iterable[str] | None"
    email_on_retry: "bool"
    email_on_failure: "bool"
    retries: "int | None"
    retry_delay: "timedelta | float"
    retry_exponential_backoff: "bool"
    max_retry_delay: "timedelta | float | None"
    start_date: "datetime | None"
    end_date: "datetime | None"
    depends_on_past: "bool"
    ignore_first_depends_on_past: "bool"
    wait_for_past_depends_before_skipping: "bool"
    wait_for_downstream: "bool"
    dag: "DAG | None"
    params: "collections.abc.MutableMapping | None"
    default_args: "dict | None"
    priority_weight: "int"
    weight_rule: "str | PriorityWeightStrategy"
    queue: "str"
    pool: "str | None"
    pool_slots: "int"
    sla: "timedelta | None"
    execution_timeout: "timedelta | None"
    on_execute_callback: (
        "None | TaskStateChangeCallback | list[TaskStateChangeCallback]"
    )
    on_failure_callback: (
        "None | TaskStateChangeCallback | list[TaskStateChangeCallback]"
    )
    on_success_callback: (
        "None | TaskStateChangeCallback | list[TaskStateChangeCallback]"
    )
    on_retry_callback: "None | TaskStateChangeCallback | list[TaskStateChangeCallback]"
    on_skipped_callback: (
        "None | TaskStateChangeCallback | list[TaskStateChangeCallback]"
    )
    pre_execute: "TaskPreExecuteHook | None"
    post_execute: "TaskPostExecuteHook | None"
    trigger_rule: "str"
    resources: "dict[str, Any] | None"
    run_as_user: "str | None"
    task_concurrency: "int | None"
    map_index_template: "str | None"
    max_active_tis_per_dag: "int | None"
    max_active_tis_per_dagrun: "int | None"
    executor: "str | None"
    executor_config: "dict | None"
    do_xcom_push: "bool"
    multiple_outputs: "bool"
    inlets: "Any | None"
    outlets: "Any | None"
    task_group: "TaskGroup | None"
    doc: "str | None"
    doc_md: "str | None"
    doc_json: "str | None"
    doc_yaml: "str | None"
    doc_rst: "str | None"
    task_display_name: "str | None"
    logger_name: "str | None"
    allow_nested_operators: "bool"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.providers.amazon.aws.operators.base_aws import (
    AwsBaseOperator,
)


class CloudFormationCreateStackOperator(AwsBaseOperator):
    stack_name: "str"
    cloudformation_parameters: "dict"


class CloudFormationDeleteStackOperator(AwsBaseOperator):
    stack_name: "str"
    cloudformation_parameters: "dict | None"
    aws_conn_id: "str | None"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.models.baseoperator import BaseOperator


class AzureFileShareToGCSOperator(BaseOperator):
    share_name: "str"
    dest_gcs: "str"
    directory_name: "str | None"
    directory_path: "str | None"
    prefix: "str"
    azure_fileshare_conn_id: "str"
    gcp_conn_id: "str"
    replace: "bool"
    gzip: "bool"
    google_impersonation_chain: "str | Sequence[str] | None"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.decorators.base import DecoratedOperator
from airfly._vendor.airflow.providers.cncf.kubernetes.operators.pod import (
    KubernetesPodOperator,
)


class _KubernetesDecoratedOperator(DecoratedOperator, KubernetesPodOperator):
    namespace: "str"
    use_dill: "bool"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.models.baseoperator import BaseOperator


class AwsToAwsBaseOperator(BaseOperator):
    source_aws_conn_id: "str | None"
    dest_aws_conn_id: "str | None | ArgNotSet"
    aws_conn_id: "str | None | ArgNotSet"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.sensors.base import BaseSensorOperator


class JenkinsBuildSensor(BaseSensorOperator):
    jenkins_connection_id: "str"
    job_name: "str"
    build_number: "int | None"
    target_states: "Iterable[str] | None"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.models.baseoperator import BaseOperator


class SFTPToWasbOperator(BaseOperator):
    sftp_source_path: "str"
    container_name: "str"
    blob_prefix: "str"
    sftp_conn_id: "str"
    wasb_conn_id: "str"
    load_options: "dict | None"
    move_object: "bool"
    wasb_overwrite_object: "bool"
    create_container: "bool"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.sensors.base import BaseSensorOperator


class MongoSensor(BaseSensorOperator):
    collection: "str"
    query: "dict"
    mongo_conn_id: "str"
    mongo_db: "_empty"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.providers.amazon.aws.sensors.base_aws import AwsBaseSensor


class GlueCatalogPartitionSensor(AwsBaseSensor):
    table_name: "str"
    expression: "str"
    database_name: "str"
    poke_interval: "int"
    deferrable: "bool"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.models.baseoperator import BaseOperator


class FlinkKubernetesOperator(BaseOperator):
    application_file: "str"
    namespace: "str | None"
    kubernetes_conn_id: "str"
    api_group: "str"
    api_version: "str"
    in_cluster: "bool | None"
    cluster_context: "str | None"
    config_file: "str | None"
    plural: "str"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.providers.amazon.aws.operators.base_aws import (
    AwsBaseOperator,
)


class QuickSightCreateIngestionOperator(AwsBaseOperator):
    data_set_id: "str"
    ingestion_id: "str"
    ingestion_type: "str"
    wait_for_completion: "bool"
    check_interval: "int"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.sensors.base import BaseSensorOperator


class JiraSensor(BaseSensorOperator):
    method_name: "str"
    jira_conn_id: "str"
    method_params: "dict | None"
    result_processor: "Callable | None"


class JiraTicketSensor(JiraSensor):
    jira_conn_id: "str"
    ticket_id: "str | None"
    field: "str | None"
    expected_value: "str | None"
    field_checker_func: "Callable | None"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.models.baseoperator import BaseOperator


class DataprocCreateClusterOperator(BaseOperator):
    folder_id: "str | None"
    cluster_name: "str | None"
    cluster_description: "str | None"
    cluster_image_version: "str | None"
    ssh_public_keys: "str | Iterable[str] | None"
    subnet_id: "str | None"
    services: "Iterable[str]"
    s3_bucket: "str | None"
    zone: "str"
    service_account_id: "str | None"
    masternode_resource_preset: "str | None"
    masternode_disk_size: "int | None"
    masternode_disk_type: "str | None"
    datanode_resource_preset: "str | None"
    datanode_disk_size: "int | None"
    datanode_disk_type: "str | None"
    datanode_count: "int"
    computenode_resource_preset: "str | None"
    computenode_disk_size: "int | None"
    computenode_disk_type: "str | None"
    computenode_count: "int"
    computenode_max_hosts_count: "int | None"
    computenode_measurement_duration: "int | None"
    computenode_warmup_duration: "int | None"
    computenode_stabilization_duration: "int | None"
    computenode_preemptible: "bool"
    computenode_cpu_utilization_target: "int | None"
    computenode_decommission_timeout: "int | None"
    connection_id: "str | None"
    properties: "dict[str, str] | None"
    enable_ui_proxy: "bool"
    host_group_ids: "Iterable[str] | None"
    security_group_ids: "Iterable[str] | None"
    log_group_id: "str | None"
    initialization_actions: "Iterable[InitializationAction] | None"
    labels: "dict[str, str] | None"


class DataprocBaseOperator(BaseOperator):
    yandex_conn_id: "str | None"
    cluster_id: "str | None"


class DataprocDeleteClusterOperator(DataprocBaseOperator):
    connection_id: "str | None"
    cluster_id: "str | None"


class DataprocCreateHiveJobOperator(DataprocBaseOperator):
    query: "str | None"
    query_file_uri: "str | None"
    script_variables: "dict[str, str] | None"
    continue_on_failure: "bool"
    properties: "dict[str, str] | None"
    name: "str"
    cluster_id: "str | None"
    connection_id: "str | None"


class DataprocCreateMapReduceJobOperator(DataprocBaseOperator):
    main_class: "str | None"
    main_jar_file_uri: "str | None"
    jar_file_uris: "Iterable[str] | None"
    archive_uris: "Iterable[str] | None"
    file_uris: "Iterable[str] | None"
    args: "Iterable[str] | None"
    properties: "dict[str, str] | None"
    name: "str"
    cluster_id: "str | None"
    connection_id: "str | None"


class DataprocCreateSparkJobOperator(DataprocBaseOperator):
    main_class: "str | None"
    main_jar_file_uri: "str | None"
    jar_file_uris: "Iterable[str] | None"
    archive_uris: "Iterable[str] | None"
    file_uris: "Iterable[str] | None"
    args: "Iterable[str] | None"
    properties: "dict[str, str] | None"
    name: "str"
    cluster_id: "str | None"
    connection_id: "str | None"
    packages: "Iterable[str] | None"
    repositories: "Iterable[str] | None"
    exclude_packages: "Iterable[str] | None"


class DataprocCreatePysparkJobOperator(DataprocBaseOperator):
    main_python_file_uri: "str | None"
    python_file_uris: "Iterable[str] | None"
    jar_file_uris: "Iterable[str] | None"
    archive_uris: "Iterable[str] | None"
    file_uris: "Iterable[str] | None"
    args: "Iterable[str] | None"
    properties: "dict[str, str] | None"
    name: "str"
    cluster_id: "str | None"
    connection_id: "str | None"
    packages: "Iterable[str] | None"
    repositories: "Iterable[str] | None"
    exclude_packages: "Iterable[str] | None"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.models.baseoperator import BaseOperator


class BigQueryToBigQueryOperator(BaseOperator):
    source_project_dataset_tables: "list[str] | str"
    destination_project_dataset_table: "str"
    write_disposition: "str"
    create_disposition: "str"
    gcp_conn_id: "str"
    labels: "dict | None"
    encryption_configuration: "dict | None"
    location: "str | None"
    impersonation_chain: "str | Sequence[str] | None"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.models.baseoperator import BaseOperator
from airfly._vendor.airflow.providers.common.sql.operators.sql import (
    SQLExecuteQueryOperator,
)


class DatabricksSqlOperator(SQLExecuteQueryOperator):
    databricks_conn_id: "str"
    http_path: "str | None"
    sql_endpoint_name: "str | None"
    session_configuration: "_empty"
    http_headers: "list[tuple[str, str]] | None"
    catalog: "str | None"
    schema: "str | None"
    output_path: "str | None"
    output_format: "str"
    csv_params: "dict[str, Any] | None"
    client_parameters: "dict[str, Any] | None"


class DatabricksCopyIntoOperator(BaseOperator):
    table_name: "str"
    file_location: "str"
    file_format: "str"
    databricks_conn_id: "str"
    http_path: "str | None"
    sql_endpoint_name: "str | None"
    session_configuration: "_empty"
    http_headers: "list[tuple[str, str]] | None"
    client_parameters: "dict[str, Any] | None"
    catalog: "str | None"
    schema: "str | None"
    files: "list[str] | None"
    pattern: "str | None"
    expression_list: "str | None"
    credential: "dict[str, str] | None"
    storage_credential: "str | None"
    encryption: "dict[str, str] | None"
    format_options: "dict[str, str] | None"
    force_copy: "bool | None"
    copy_options: "dict[str, str] | None"
    validate: "bool | int | None"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.models.baseoperator import BaseOperator


class GCSToSFTPOperator(BaseOperator):
    source_bucket: "str"
    source_object: "str"
    destination_path: "str"
    keep_directory_structure: "bool"
    move_object: "bool"
    gcp_conn_id: "str"
    sftp_conn_id: "str"
    impersonation_chain: "str | Sequence[str] | None"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.operators.branch import BaseBranchOperator


class BranchDayOfWeekOperator(BaseBranchOperator):
    follow_task_ids_if_true: "str | Iterable[str]"
    follow_task_ids_if_false: "str | Iterable[str]"
    week_day: "str | Iterable[str] | WeekDay | Iterable[WeekDay]"
    use_task_logical_date: "bool"
    use_task_execution_day: "bool"



# Auto generated by 'inv collect-airflow'
from airfly._vendor.airflow.providers.google.cloud.operators.cloud_base import (
    GoogleCloudBaseOperator,
)


class CloudSpeechToTextRecognizeSpeechOperator(GoogleCloudBaseOperator):
    audio: "RecognitionAudio"
    config: "RecognitionConfig"
    project_id: "str"
    gcp_conn_id: "str"
    retry: "Retry | _MethodDefault"
    timeout: "float | None"
    impersonation_chain: "str | Sequence[str] | None"
